/*
 * Copyright (c) 2019-2020 Arm Limited.
 *
 * SPDX-License-Identifier: MIT
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */
#ifdef __ARM_FEATURE_SVE

#include "arm_gemm.hpp"
#include "../../utils.hpp"

#include <cassert>

namespace arm_gemm {

void sve_hybrid_s8s32_dot_6x4VL (
    unsigned int num_strings, const unsigned int *string_lengths, IndirectInputArg<int8_t> A_arg,
    size_t M, size_t N, const int8_t *B_ptr, IndirectOutputArg<int32_t> output_arg,
    const int32_t *, Activation, bool accumulate
)
{
    struct KernelArgs {
        unsigned int num_strings = {};
        const unsigned int *string_lengths = {};
        size_t N = {};
        const int8_t *B_ptr = {};
        size_t output_offset = {};
        size_t input_initial_col = {};
        size_t input_offset = {};
    } ka;

    unsigned long flags=0;
    void *output_ptr;
    void *input_ptr;

    if (output_arg.is_indirect) {
        output_ptr=(void *)(output_arg.indirect.ptr);
        ka.output_offset=output_arg.indirect.offset;
        flags |= 0x4;
    } else {
        output_ptr=(void *)(output_arg.direct.base);
        ka.output_offset=output_arg.direct.stride;
    }

    if (A_arg.is_indirect) {
        input_ptr=(void *)(A_arg.indirect.ptr);
        ka.input_offset=A_arg.indirect.start_row;
        ka.input_initial_col=A_arg.indirect.start_col;
        flags |= 0x8;
    } else {
        assert(num_strings==1);
        input_ptr=(void *)(A_arg.direct.base);
        ka.input_offset=A_arg.direct.stride;
    }
    if (accumulate) {
        flags |= 0x1;
    }
    ka.num_strings = num_strings;
    ka.string_lengths = string_lengths;
    ka.N = N;
    ka.B_ptr = B_ptr;
    __asm__ __volatile__(
      "ptrue p5.b\n"
      "1:"  // Row loop
      "cmp %x[M], #0x6\n"
      "bge 61f\n"
      "cmp %x[M], #0x4\n"
      "bgt 49f\n"
      "beq 37f\n"
      "cmp %x[M], #0x2\n"
      "bgt 25f\n"
      "beq 13f\n"
      "ldr x15, [%x[args_ptr], %[offsetof_N]]\n"
      "ldr x14, [%x[args_ptr], %[offsetof_B_ptr]]\n"
      "ldr x19, [%x[args_ptr], %[offsetof_output_offset]]\n"
      "tbz %x[flags], #2, 2f\n"
      "ldr x13, [%x[output_ptr], #0x0]\n"
      "add x13, x13, x19, LSL #2\n"
      "b 3f\n"
      "2:"  // Height 1: setup direct output
      "mov x13, %x[output_ptr]\n"
      "3:"  // Height 1: Column loop
      "mov x19, #0x0\n"
      "whilelt p4.s, x19, x15\n"
      "incw x19\n"
      "whilelt p3.s, x19, x15\n"
      "incw x19\n"
      "whilelt p2.s, x19, x15\n"
      "incw x19\n"
      "whilelt p1.s, x19, x15\n"
      "tbz %x[flags], #0, 4f\n"
      "ld1w { z8.s }, p4/Z, [x13]\n"
      "ld1w { z9.s }, p3/Z, [x13, #1, MUL VL]\n"
      "ld1w { z10.s }, p2/Z, [x13, #2, MUL VL]\n"
      "ld1w { z11.s }, p1/Z, [x13, #3, MUL VL]\n"
      "b 5f\n"
      "4:"  // Height 1: no accumulate
      "mov z8.s, #0x0\n"
      "mov z9.s, #0x0\n"
      "mov z10.s, #0x0\n"
      "mov z11.s, #0x0\n"
      "5:"  // Height 1: setup done
      "mov x12, #0x0\n"
      "6:"  // Height 1: String loop
      "ldr x20, [%x[args_ptr], %[offsetof_string_lengths]]\n"
      "ldr x19, [%x[args_ptr], %[offsetof_input_offset]]\n"
      "ldr w11, [x20, x12, LSL #0x2]\n"
      "tbz %x[flags], #3, 7f\n"
      "ldr x20, [%x[input_ptr], x12, LSL #0x3]\n"
      "add x20, x20, x19, LSL #3\n"
      "ldr x10, [x20, #0x0]\n"
      "cbnz x12, 8f\n"
      "ldr x19, [%x[args_ptr], %[offsetof_input_initial_col]]\n"
      "add x10, x10, x19\n"
      "b 8f\n"
      "7:"  // Height 1: setup direct input
      "mov x10, %x[input_ptr]\n"
      "8:"  // Height 1: input setup done
      "cmp x11, #0x10\n"
      "ble 10f\n"
      "9:"  // Height 1: Multiply loop: Main loop head
      "ld1b { z6.b }, p5/Z, [x14]\n"
      "whilelt p0.b, XZR, x11\n"
      "ld1b { z7.b }, p5/Z, [x14, #1, MUL VL]\n"
      "sub x11, x11, #0x10\n"
      "ld1rqb { z0.b }, p0/Z, [x10]\n"
      "sdot z8.s, z6.b, z0.b[0]\n"
      "ld1b { z6.b }, p5/Z, [x14, #2, MUL VL]\n"
      "add x10, x10, #0x10\n"
      "sdot z9.s, z7.b, z0.b[0]\n"
      "ld1b { z7.b }, p5/Z, [x14, #3, MUL VL]\n"
      "cmp x11, #0x10\n"
      "sdot z10.s, z6.b, z0.b[0]\n"
      "ld1b { z6.b }, p5/Z, [x14, #4, MUL VL]\n"
      "prfm pldl1keep, [x10, #0x80]\n"
      "sdot z11.s, z7.b, z0.b[0]\n"
      "ld1b { z7.b }, p5/Z, [x14, #5, MUL VL]\n"
      "sdot z8.s, z6.b, z0.b[1]\n"
      "ld1b { z6.b }, p5/Z, [x14, #6, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[1]\n"
      "ld1b { z7.b }, p5/Z, [x14, #7, MUL VL]\n"
      "addvl x14, x14, #16\n"
      "sdot z10.s, z6.b, z0.b[1]\n"
      "ld1b { z6.b }, p5/Z, [x14, #-8, MUL VL]\n"
      "sdot z11.s, z7.b, z0.b[1]\n"
      "ld1b { z7.b }, p5/Z, [x14, #-7, MUL VL]\n"
      "sdot z8.s, z6.b, z0.b[2]\n"
      "ld1b { z6.b }, p5/Z, [x14, #-6, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[2]\n"
      "ld1b { z7.b }, p5/Z, [x14, #-5, MUL VL]\n"
      "sdot z10.s, z6.b, z0.b[2]\n"
      "ld1b { z6.b }, p5/Z, [x14, #-4, MUL VL]\n"
      "sdot z11.s, z7.b, z0.b[2]\n"
      "ld1b { z7.b }, p5/Z, [x14, #-3, MUL VL]\n"
      "sdot z8.s, z6.b, z0.b[3]\n"
      "ld1b { z6.b }, p5/Z, [x14, #-2, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[3]\n"
      "ld1b { z7.b }, p5/Z, [x14, #-1, MUL VL]\n"
      "sdot z10.s, z6.b, z0.b[3]\n"
      "sdot z11.s, z7.b, z0.b[3]\n"
      "bgt 9b\n"
      "10:"  // Height 1: Multiply loop: Single iteration only
      "ld1b { z6.b }, p5/Z, [x14]\n"
      "whilelt p0.b, XZR, x11\n"
      "ld1b { z7.b }, p5/Z, [x14, #1, MUL VL]\n"
      "subs x11, x11, #0x4\n"
      "ld1rqb { z0.b }, p0/Z, [x10]\n"
      "sdot z8.s, z6.b, z0.b[0]\n"
      "ld1b { z6.b }, p5/Z, [x14, #2, MUL VL]\n"
      "add x10, x10, #0x10\n"
      "sdot z9.s, z7.b, z0.b[0]\n"
      "ld1b { z7.b }, p5/Z, [x14, #3, MUL VL]\n"
      "addvl x14, x14, #4\n"
      "sdot z10.s, z6.b, z0.b[0]\n"
      "sdot z11.s, z7.b, z0.b[0]\n"
      "ble 11f\n"
      "ld1b { z6.b }, p5/Z, [x14]\n"
      "sdot z8.s, z6.b, z0.b[1]\n"
      "ld1b { z7.b }, p5/Z, [x14, #1, MUL VL]\n"
      "subs x11, x11, #0x4\n"
      "sdot z9.s, z7.b, z0.b[1]\n"
      "ld1b { z6.b }, p5/Z, [x14, #2, MUL VL]\n"
      "ld1b { z7.b }, p5/Z, [x14, #3, MUL VL]\n"
      "sdot z10.s, z6.b, z0.b[1]\n"
      "addvl x14, x14, #4\n"
      "sdot z11.s, z7.b, z0.b[1]\n"
      "ble 11f\n"
      "ld1b { z6.b }, p5/Z, [x14]\n"
      "sdot z8.s, z6.b, z0.b[2]\n"
      "ld1b { z7.b }, p5/Z, [x14, #1, MUL VL]\n"
      "subs x11, x11, #0x4\n"
      "sdot z9.s, z7.b, z0.b[2]\n"
      "ld1b { z6.b }, p5/Z, [x14, #2, MUL VL]\n"
      "ld1b { z7.b }, p5/Z, [x14, #3, MUL VL]\n"
      "sdot z10.s, z6.b, z0.b[2]\n"
      "addvl x14, x14, #4\n"
      "sdot z11.s, z7.b, z0.b[2]\n"
      "ble 11f\n"
      "ld1b { z6.b }, p5/Z, [x14]\n"
      "sdot z8.s, z6.b, z0.b[3]\n"
      "ld1b { z7.b }, p5/Z, [x14, #1, MUL VL]\n"
      "ld1b { z6.b }, p5/Z, [x14, #2, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[3]\n"
      "ld1b { z7.b }, p5/Z, [x14, #3, MUL VL]\n"
      "addvl x14, x14, #4\n"
      "sdot z10.s, z6.b, z0.b[3]\n"
      "sdot z11.s, z7.b, z0.b[3]\n"
      "11:"  // Height 1: Multiply loop: multiply skip
      "prfm pldl1keep, [x10, #0x80]\n"
      "add x12, x12, #0x1\n"
      "ldr w19, [%x[args_ptr], %[offsetof_num_strings]]\n"
      "cmp x12, x19\n"
      "bne 6b\n"
      "prfm pstl1keep, [x13, #0x0]\n"
      "st1w { z8.s }, p4, [x13]\n"
      "st1w { z9.s }, p3, [x13, #1, MUL VL]\n"
      "st1w { z10.s }, p2, [x13, #2, MUL VL]\n"
      "st1w { z11.s }, p1, [x13, #3, MUL VL]\n"
      "addvl x13, x13, #4\n"
      "12:"  // Height 1: Writeback done
      "mov x19, #0x0\n"
      "incw x19, ALL, MUL #4\n"
      "subs x15, x15, x19\n"
      "bgt 3b\n"
      "b 74f\n"
      "13:"  // Height 2
      "ldr x15, [%x[args_ptr], %[offsetof_N]]\n"
      "ldr x14, [%x[args_ptr], %[offsetof_B_ptr]]\n"
      "ldr x19, [%x[args_ptr], %[offsetof_output_offset]]\n"
      "tbz %x[flags], #2, 14f\n"
      "ldr x13, [%x[output_ptr], #0x0]\n"
      "add x13, x13, x19, LSL #2\n"
      "ldr x9, [%x[output_ptr], #0x8]\n"
      "add x9, x9, x19, LSL #2\n"
      "b 15f\n"
      "14:"  // Height 2: setup direct output
      "mov x13, %x[output_ptr]\n"
      "add x9, x13, x19, LSL #2\n"
      "15:"  // Height 2: Column loop
      "mov x19, #0x0\n"
      "whilelt p4.s, x19, x15\n"
      "incw x19\n"
      "whilelt p3.s, x19, x15\n"
      "incw x19\n"
      "whilelt p2.s, x19, x15\n"
      "incw x19\n"
      "whilelt p1.s, x19, x15\n"
      "tbz %x[flags], #0, 16f\n"
      "ld1w { z8.s }, p4/Z, [x13]\n"
      "ld1w { z9.s }, p3/Z, [x13, #1, MUL VL]\n"
      "ld1w { z10.s }, p2/Z, [x13, #2, MUL VL]\n"
      "ld1w { z11.s }, p1/Z, [x13, #3, MUL VL]\n"
      "ld1w { z12.s }, p4/Z, [x9]\n"
      "ld1w { z13.s }, p3/Z, [x9, #1, MUL VL]\n"
      "ld1w { z14.s }, p2/Z, [x9, #2, MUL VL]\n"
      "ld1w { z15.s }, p1/Z, [x9, #3, MUL VL]\n"
      "b 17f\n"
      "16:"  // Height 2: no accumulate
      "mov z8.s, #0x0\n"
      "mov z9.s, #0x0\n"
      "mov z10.s, #0x0\n"
      "mov z11.s, #0x0\n"
      "mov z12.s, #0x0\n"
      "mov z13.s, #0x0\n"
      "mov z14.s, #0x0\n"
      "mov z15.s, #0x0\n"
      "17:"  // Height 2: setup done
      "mov x12, #0x0\n"
      "18:"  // Height 2: String loop
      "ldr x20, [%x[args_ptr], %[offsetof_string_lengths]]\n"
      "ldr x19, [%x[args_ptr], %[offsetof_input_offset]]\n"
      "ldr w11, [x20, x12, LSL #0x2]\n"
      "tbz %x[flags], #3, 19f\n"
      "ldr x20, [%x[input_ptr], x12, LSL #0x3]\n"
      "add x20, x20, x19, LSL #3\n"
      "ldr x10, [x20, #0x0]\n"
      "ldr x28, [x20, #0x8]\n"
      "cbnz x12, 20f\n"
      "ldr x19, [%x[args_ptr], %[offsetof_input_initial_col]]\n"
      "add x10, x10, x19\n"
      "add x28, x28, x19\n"
      "b 20f\n"
      "19:"  // Height 2: setup direct input
      "mov x10, %x[input_ptr]\n"
      "add x28, x10, x19\n"
      "20:"  // Height 2: input setup done
      "cmp x11, #0x10\n"
      "ble 22f\n"
      "21:"  // Height 2: Multiply loop: Main loop head
      "ld1b { z6.b }, p5/Z, [x14]\n"
      "whilelt p0.b, XZR, x11\n"
      "ld1b { z7.b }, p5/Z, [x14, #1, MUL VL]\n"
      "sub x11, x11, #0x10\n"
      "ld1rqb { z0.b }, p0/Z, [x10]\n"
      "sdot z8.s, z6.b, z0.b[0]\n"
      "ld1rqb { z1.b }, p0/Z, [x28]\n"
      "add x10, x10, #0x10\n"
      "sdot z9.s, z7.b, z0.b[0]\n"
      "prfm pldl1keep, [x10, #0x80]\n"
      "add x28, x28, #0x10\n"
      "sdot z12.s, z6.b, z1.b[0]\n"
      "ld1b { z6.b }, p5/Z, [x14, #2, MUL VL]\n"
      "cmp x11, #0x10\n"
      "sdot z13.s, z7.b, z1.b[0]\n"
      "ld1b { z7.b }, p5/Z, [x14, #3, MUL VL]\n"
      "prfm pldl1keep, [x28, #0x80]\n"
      "sdot z10.s, z6.b, z0.b[0]\n"
      "sdot z14.s, z6.b, z1.b[0]\n"
      "ld1b { z6.b }, p5/Z, [x14, #4, MUL VL]\n"
      "sdot z11.s, z7.b, z0.b[0]\n"
      "sdot z15.s, z7.b, z1.b[0]\n"
      "ld1b { z7.b }, p5/Z, [x14, #5, MUL VL]\n"
      "sdot z8.s, z6.b, z0.b[1]\n"
      "sdot z12.s, z6.b, z1.b[1]\n"
      "ld1b { z6.b }, p5/Z, [x14, #6, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[1]\n"
      "sdot z13.s, z7.b, z1.b[1]\n"
      "ld1b { z7.b }, p5/Z, [x14, #7, MUL VL]\n"
      "addvl x14, x14, #16\n"
      "sdot z10.s, z6.b, z0.b[1]\n"
      "sdot z14.s, z6.b, z1.b[1]\n"
      "ld1b { z6.b }, p5/Z, [x14, #-8, MUL VL]\n"
      "sdot z11.s, z7.b, z0.b[1]\n"
      "sdot z15.s, z7.b, z1.b[1]\n"
      "ld1b { z7.b }, p5/Z, [x14, #-7, MUL VL]\n"
      "sdot z8.s, z6.b, z0.b[2]\n"
      "sdot z12.s, z6.b, z1.b[2]\n"
      "ld1b { z6.b }, p5/Z, [x14, #-6, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[2]\n"
      "sdot z13.s, z7.b, z1.b[2]\n"
      "ld1b { z7.b }, p5/Z, [x14, #-5, MUL VL]\n"
      "sdot z10.s, z6.b, z0.b[2]\n"
      "sdot z14.s, z6.b, z1.b[2]\n"
      "ld1b { z6.b }, p5/Z, [x14, #-4, MUL VL]\n"
      "sdot z11.s, z7.b, z0.b[2]\n"
      "sdot z15.s, z7.b, z1.b[2]\n"
      "ld1b { z7.b }, p5/Z, [x14, #-3, MUL VL]\n"
      "sdot z8.s, z6.b, z0.b[3]\n"
      "sdot z12.s, z6.b, z1.b[3]\n"
      "ld1b { z6.b }, p5/Z, [x14, #-2, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[3]\n"
      "sdot z13.s, z7.b, z1.b[3]\n"
      "ld1b { z7.b }, p5/Z, [x14, #-1, MUL VL]\n"
      "sdot z10.s, z6.b, z0.b[3]\n"
      "sdot z14.s, z6.b, z1.b[3]\n"
      "sdot z11.s, z7.b, z0.b[3]\n"
      "sdot z15.s, z7.b, z1.b[3]\n"
      "bgt 21b\n"
      "22:"  // Height 2: Multiply loop: Single iteration only
      "ld1b { z6.b }, p5/Z, [x14]\n"
      "whilelt p0.b, XZR, x11\n"
      "ld1b { z7.b }, p5/Z, [x14, #1, MUL VL]\n"
      "subs x11, x11, #0x4\n"
      "ld1rqb { z0.b }, p0/Z, [x10]\n"
      "sdot z8.s, z6.b, z0.b[0]\n"
      "ld1rqb { z1.b }, p0/Z, [x28]\n"
      "add x10, x10, #0x10\n"
      "sdot z9.s, z7.b, z0.b[0]\n"
      "add x28, x28, #0x10\n"
      "sdot z12.s, z6.b, z1.b[0]\n"
      "ld1b { z6.b }, p5/Z, [x14, #2, MUL VL]\n"
      "sdot z13.s, z7.b, z1.b[0]\n"
      "ld1b { z7.b }, p5/Z, [x14, #3, MUL VL]\n"
      "addvl x14, x14, #4\n"
      "sdot z10.s, z6.b, z0.b[0]\n"
      "sdot z14.s, z6.b, z1.b[0]\n"
      "sdot z11.s, z7.b, z0.b[0]\n"
      "sdot z15.s, z7.b, z1.b[0]\n"
      "ble 23f\n"
      "ld1b { z6.b }, p5/Z, [x14]\n"
      "sdot z8.s, z6.b, z0.b[1]\n"
      "ld1b { z7.b }, p5/Z, [x14, #1, MUL VL]\n"
      "subs x11, x11, #0x4\n"
      "sdot z12.s, z6.b, z1.b[1]\n"
      "ld1b { z6.b }, p5/Z, [x14, #2, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[1]\n"
      "sdot z13.s, z7.b, z1.b[1]\n"
      "ld1b { z7.b }, p5/Z, [x14, #3, MUL VL]\n"
      "addvl x14, x14, #4\n"
      "sdot z10.s, z6.b, z0.b[1]\n"
      "sdot z14.s, z6.b, z1.b[1]\n"
      "sdot z11.s, z7.b, z0.b[1]\n"
      "sdot z15.s, z7.b, z1.b[1]\n"
      "ble 23f\n"
      "ld1b { z6.b }, p5/Z, [x14]\n"
      "sdot z8.s, z6.b, z0.b[2]\n"
      "ld1b { z7.b }, p5/Z, [x14, #1, MUL VL]\n"
      "subs x11, x11, #0x4\n"
      "sdot z12.s, z6.b, z1.b[2]\n"
      "ld1b { z6.b }, p5/Z, [x14, #2, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[2]\n"
      "sdot z13.s, z7.b, z1.b[2]\n"
      "ld1b { z7.b }, p5/Z, [x14, #3, MUL VL]\n"
      "addvl x14, x14, #4\n"
      "sdot z10.s, z6.b, z0.b[2]\n"
      "sdot z14.s, z6.b, z1.b[2]\n"
      "sdot z11.s, z7.b, z0.b[2]\n"
      "sdot z15.s, z7.b, z1.b[2]\n"
      "ble 23f\n"
      "ld1b { z6.b }, p5/Z, [x14]\n"
      "sdot z8.s, z6.b, z0.b[3]\n"
      "ld1b { z7.b }, p5/Z, [x14, #1, MUL VL]\n"
      "sdot z12.s, z6.b, z1.b[3]\n"
      "ld1b { z6.b }, p5/Z, [x14, #2, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[3]\n"
      "sdot z13.s, z7.b, z1.b[3]\n"
      "ld1b { z7.b }, p5/Z, [x14, #3, MUL VL]\n"
      "addvl x14, x14, #4\n"
      "sdot z10.s, z6.b, z0.b[3]\n"
      "sdot z14.s, z6.b, z1.b[3]\n"
      "sdot z11.s, z7.b, z0.b[3]\n"
      "sdot z15.s, z7.b, z1.b[3]\n"
      "23:"  // Height 2: Multiply loop: multiply skip
      "prfm pldl1keep, [x10, #0x80]\n"
      "add x12, x12, #0x1\n"
      "prfm pldl1keep, [x28, #0x80]\n"
      "ldr w19, [%x[args_ptr], %[offsetof_num_strings]]\n"
      "cmp x12, x19\n"
      "bne 18b\n"
      "prfm pstl1keep, [x13, #0x0]\n"
      "prfm pstl1keep, [x9, #0x0]\n"
      "st1w { z8.s }, p4, [x13]\n"
      "st1w { z9.s }, p3, [x13, #1, MUL VL]\n"
      "st1w { z10.s }, p2, [x13, #2, MUL VL]\n"
      "st1w { z11.s }, p1, [x13, #3, MUL VL]\n"
      "addvl x13, x13, #4\n"
      "st1w { z12.s }, p4, [x9]\n"
      "st1w { z13.s }, p3, [x9, #1, MUL VL]\n"
      "st1w { z14.s }, p2, [x9, #2, MUL VL]\n"
      "st1w { z15.s }, p1, [x9, #3, MUL VL]\n"
      "addvl x9, x9, #4\n"
      "24:"  // Height 2: Writeback done
      "mov x19, #0x0\n"
      "incw x19, ALL, MUL #4\n"
      "subs x15, x15, x19\n"
      "bgt 15b\n"
      "b 74f\n"
      "25:"  // Height 3
      "ldr x15, [%x[args_ptr], %[offsetof_N]]\n"
      "ldr x14, [%x[args_ptr], %[offsetof_B_ptr]]\n"
      "ldr x19, [%x[args_ptr], %[offsetof_output_offset]]\n"
      "tbz %x[flags], #2, 26f\n"
      "ldr x13, [%x[output_ptr], #0x0]\n"
      "add x13, x13, x19, LSL #2\n"
      "ldr x9, [%x[output_ptr], #0x8]\n"
      "ldr x27, [%x[output_ptr], #0x10]\n"
      "add x9, x9, x19, LSL #2\n"
      "add x27, x27, x19, LSL #2\n"
      "b 27f\n"
      "26:"  // Height 3: setup direct output
      "mov x13, %x[output_ptr]\n"
      "add x9, x13, x19, LSL #2\n"
      "add x27, x9, x19, LSL #2\n"
      "27:"  // Height 3: Column loop
      "mov x19, #0x0\n"
      "whilelt p4.s, x19, x15\n"
      "incw x19\n"
      "whilelt p3.s, x19, x15\n"
      "incw x19\n"
      "whilelt p2.s, x19, x15\n"
      "incw x19\n"
      "whilelt p1.s, x19, x15\n"
      "tbz %x[flags], #0, 28f\n"
      "ld1w { z8.s }, p4/Z, [x13]\n"
      "ld1w { z9.s }, p3/Z, [x13, #1, MUL VL]\n"
      "ld1w { z10.s }, p2/Z, [x13, #2, MUL VL]\n"
      "ld1w { z11.s }, p1/Z, [x13, #3, MUL VL]\n"
      "ld1w { z12.s }, p4/Z, [x9]\n"
      "ld1w { z13.s }, p3/Z, [x9, #1, MUL VL]\n"
      "ld1w { z14.s }, p2/Z, [x9, #2, MUL VL]\n"
      "ld1w { z15.s }, p1/Z, [x9, #3, MUL VL]\n"
      "ld1w { z16.s }, p4/Z, [x27]\n"
      "ld1w { z17.s }, p3/Z, [x27, #1, MUL VL]\n"
      "ld1w { z18.s }, p2/Z, [x27, #2, MUL VL]\n"
      "ld1w { z19.s }, p1/Z, [x27, #3, MUL VL]\n"
      "b 29f\n"
      "28:"  // Height 3: no accumulate
      "mov z8.s, #0x0\n"
      "mov z9.s, #0x0\n"
      "mov z10.s, #0x0\n"
      "mov z11.s, #0x0\n"
      "mov z12.s, #0x0\n"
      "mov z13.s, #0x0\n"
      "mov z14.s, #0x0\n"
      "mov z15.s, #0x0\n"
      "mov z16.s, #0x0\n"
      "mov z17.s, #0x0\n"
      "mov z18.s, #0x0\n"
      "mov z19.s, #0x0\n"
      "29:"  // Height 3: setup done
      "mov x12, #0x0\n"
      "30:"  // Height 3: String loop
      "ldr x20, [%x[args_ptr], %[offsetof_string_lengths]]\n"
      "ldr x19, [%x[args_ptr], %[offsetof_input_offset]]\n"
      "ldr w11, [x20, x12, LSL #0x2]\n"
      "tbz %x[flags], #3, 31f\n"
      "ldr x20, [%x[input_ptr], x12, LSL #0x3]\n"
      "add x20, x20, x19, LSL #3\n"
      "ldr x10, [x20, #0x0]\n"
      "ldr x28, [x20, #0x8]\n"
      "ldr x26, [x20, #0x10]\n"
      "cbnz x12, 32f\n"
      "ldr x19, [%x[args_ptr], %[offsetof_input_initial_col]]\n"
      "add x10, x10, x19\n"
      "add x28, x28, x19\n"
      "add x26, x26, x19\n"
      "b 32f\n"
      "31:"  // Height 3: setup direct input
      "mov x10, %x[input_ptr]\n"
      "add x28, x10, x19\n"
      "add x26, x28, x19\n"
      "32:"  // Height 3: input setup done
      "cmp x11, #0x10\n"
      "ble 34f\n"
      "33:"  // Height 3: Multiply loop: Main loop head
      "ld1b { z6.b }, p5/Z, [x14]\n"
      "whilelt p0.b, XZR, x11\n"
      "ld1b { z7.b }, p5/Z, [x14, #1, MUL VL]\n"
      "sub x11, x11, #0x10\n"
      "ld1rqb { z0.b }, p0/Z, [x10]\n"
      "sdot z8.s, z6.b, z0.b[0]\n"
      "ld1rqb { z1.b }, p0/Z, [x28]\n"
      "add x10, x10, #0x10\n"
      "sdot z9.s, z7.b, z0.b[0]\n"
      "ld1rqb { z2.b }, p0/Z, [x26]\n"
      "add x28, x28, #0x10\n"
      "sdot z12.s, z6.b, z1.b[0]\n"
      "prfm pldl1keep, [x10, #0x80]\n"
      "add x26, x26, #0x10\n"
      "sdot z16.s, z6.b, z2.b[0]\n"
      "ld1b { z6.b }, p5/Z, [x14, #2, MUL VL]\n"
      "cmp x11, #0x10\n"
      "sdot z13.s, z7.b, z1.b[0]\n"
      "prfm pldl1keep, [x28, #0x80]\n"
      "sdot z17.s, z7.b, z2.b[0]\n"
      "ld1b { z7.b }, p5/Z, [x14, #3, MUL VL]\n"
      "prfm pldl1keep, [x26, #0x80]\n"
      "sdot z10.s, z6.b, z0.b[0]\n"
      "sdot z14.s, z6.b, z1.b[0]\n"
      "sdot z18.s, z6.b, z2.b[0]\n"
      "ld1b { z6.b }, p5/Z, [x14, #4, MUL VL]\n"
      "sdot z11.s, z7.b, z0.b[0]\n"
      "sdot z15.s, z7.b, z1.b[0]\n"
      "sdot z19.s, z7.b, z2.b[0]\n"
      "ld1b { z7.b }, p5/Z, [x14, #5, MUL VL]\n"
      "sdot z8.s, z6.b, z0.b[1]\n"
      "sdot z12.s, z6.b, z1.b[1]\n"
      "sdot z16.s, z6.b, z2.b[1]\n"
      "ld1b { z6.b }, p5/Z, [x14, #6, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[1]\n"
      "sdot z13.s, z7.b, z1.b[1]\n"
      "sdot z17.s, z7.b, z2.b[1]\n"
      "ld1b { z7.b }, p5/Z, [x14, #7, MUL VL]\n"
      "addvl x14, x14, #16\n"
      "sdot z10.s, z6.b, z0.b[1]\n"
      "sdot z14.s, z6.b, z1.b[1]\n"
      "sdot z18.s, z6.b, z2.b[1]\n"
      "ld1b { z6.b }, p5/Z, [x14, #-8, MUL VL]\n"
      "sdot z11.s, z7.b, z0.b[1]\n"
      "sdot z15.s, z7.b, z1.b[1]\n"
      "sdot z19.s, z7.b, z2.b[1]\n"
      "ld1b { z7.b }, p5/Z, [x14, #-7, MUL VL]\n"
      "sdot z8.s, z6.b, z0.b[2]\n"
      "sdot z12.s, z6.b, z1.b[2]\n"
      "sdot z16.s, z6.b, z2.b[2]\n"
      "ld1b { z6.b }, p5/Z, [x14, #-6, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[2]\n"
      "sdot z13.s, z7.b, z1.b[2]\n"
      "sdot z17.s, z7.b, z2.b[2]\n"
      "ld1b { z7.b }, p5/Z, [x14, #-5, MUL VL]\n"
      "sdot z10.s, z6.b, z0.b[2]\n"
      "sdot z14.s, z6.b, z1.b[2]\n"
      "sdot z18.s, z6.b, z2.b[2]\n"
      "ld1b { z6.b }, p5/Z, [x14, #-4, MUL VL]\n"
      "sdot z11.s, z7.b, z0.b[2]\n"
      "sdot z15.s, z7.b, z1.b[2]\n"
      "sdot z19.s, z7.b, z2.b[2]\n"
      "ld1b { z7.b }, p5/Z, [x14, #-3, MUL VL]\n"
      "sdot z8.s, z6.b, z0.b[3]\n"
      "sdot z12.s, z6.b, z1.b[3]\n"
      "sdot z16.s, z6.b, z2.b[3]\n"
      "ld1b { z6.b }, p5/Z, [x14, #-2, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[3]\n"
      "sdot z13.s, z7.b, z1.b[3]\n"
      "sdot z17.s, z7.b, z2.b[3]\n"
      "ld1b { z7.b }, p5/Z, [x14, #-1, MUL VL]\n"
      "sdot z10.s, z6.b, z0.b[3]\n"
      "sdot z14.s, z6.b, z1.b[3]\n"
      "sdot z18.s, z6.b, z2.b[3]\n"
      "sdot z11.s, z7.b, z0.b[3]\n"
      "sdot z15.s, z7.b, z1.b[3]\n"
      "sdot z19.s, z7.b, z2.b[3]\n"
      "bgt 33b\n"
      "34:"  // Height 3: Multiply loop: Single iteration only
      "ld1b { z6.b }, p5/Z, [x14]\n"
      "whilelt p0.b, XZR, x11\n"
      "ld1b { z7.b }, p5/Z, [x14, #1, MUL VL]\n"
      "subs x11, x11, #0x4\n"
      "ld1rqb { z0.b }, p0/Z, [x10]\n"
      "sdot z8.s, z6.b, z0.b[0]\n"
      "ld1rqb { z1.b }, p0/Z, [x28]\n"
      "add x10, x10, #0x10\n"
      "sdot z9.s, z7.b, z0.b[0]\n"
      "ld1rqb { z2.b }, p0/Z, [x26]\n"
      "add x28, x28, #0x10\n"
      "sdot z12.s, z6.b, z1.b[0]\n"
      "add x26, x26, #0x10\n"
      "sdot z13.s, z7.b, z1.b[0]\n"
      "sdot z16.s, z6.b, z2.b[0]\n"
      "ld1b { z6.b }, p5/Z, [x14, #2, MUL VL]\n"
      "sdot z17.s, z7.b, z2.b[0]\n"
      "ld1b { z7.b }, p5/Z, [x14, #3, MUL VL]\n"
      "addvl x14, x14, #4\n"
      "sdot z10.s, z6.b, z0.b[0]\n"
      "sdot z14.s, z6.b, z1.b[0]\n"
      "sdot z18.s, z6.b, z2.b[0]\n"
      "sdot z11.s, z7.b, z0.b[0]\n"
      "sdot z15.s, z7.b, z1.b[0]\n"
      "sdot z19.s, z7.b, z2.b[0]\n"
      "ble 35f\n"
      "ld1b { z6.b }, p5/Z, [x14]\n"
      "sdot z8.s, z6.b, z0.b[1]\n"
      "ld1b { z7.b }, p5/Z, [x14, #1, MUL VL]\n"
      "subs x11, x11, #0x4\n"
      "sdot z12.s, z6.b, z1.b[1]\n"
      "sdot z16.s, z6.b, z2.b[1]\n"
      "ld1b { z6.b }, p5/Z, [x14, #2, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[1]\n"
      "sdot z13.s, z7.b, z1.b[1]\n"
      "sdot z17.s, z7.b, z2.b[1]\n"
      "ld1b { z7.b }, p5/Z, [x14, #3, MUL VL]\n"
      "addvl x14, x14, #4\n"
      "sdot z10.s, z6.b, z0.b[1]\n"
      "sdot z14.s, z6.b, z1.b[1]\n"
      "sdot z18.s, z6.b, z2.b[1]\n"
      "sdot z11.s, z7.b, z0.b[1]\n"
      "sdot z15.s, z7.b, z1.b[1]\n"
      "sdot z19.s, z7.b, z2.b[1]\n"
      "ble 35f\n"
      "ld1b { z6.b }, p5/Z, [x14]\n"
      "sdot z8.s, z6.b, z0.b[2]\n"
      "ld1b { z7.b }, p5/Z, [x14, #1, MUL VL]\n"
      "subs x11, x11, #0x4\n"
      "sdot z12.s, z6.b, z1.b[2]\n"
      "sdot z16.s, z6.b, z2.b[2]\n"
      "ld1b { z6.b }, p5/Z, [x14, #2, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[2]\n"
      "sdot z13.s, z7.b, z1.b[2]\n"
      "sdot z17.s, z7.b, z2.b[2]\n"
      "ld1b { z7.b }, p5/Z, [x14, #3, MUL VL]\n"
      "addvl x14, x14, #4\n"
      "sdot z10.s, z6.b, z0.b[2]\n"
      "sdot z14.s, z6.b, z1.b[2]\n"
      "sdot z18.s, z6.b, z2.b[2]\n"
      "sdot z11.s, z7.b, z0.b[2]\n"
      "sdot z15.s, z7.b, z1.b[2]\n"
      "sdot z19.s, z7.b, z2.b[2]\n"
      "ble 35f\n"
      "ld1b { z6.b }, p5/Z, [x14]\n"
      "sdot z8.s, z6.b, z0.b[3]\n"
      "ld1b { z7.b }, p5/Z, [x14, #1, MUL VL]\n"
      "sdot z12.s, z6.b, z1.b[3]\n"
      "sdot z16.s, z6.b, z2.b[3]\n"
      "ld1b { z6.b }, p5/Z, [x14, #2, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[3]\n"
      "sdot z13.s, z7.b, z1.b[3]\n"
      "sdot z17.s, z7.b, z2.b[3]\n"
      "ld1b { z7.b }, p5/Z, [x14, #3, MUL VL]\n"
      "addvl x14, x14, #4\n"
      "sdot z10.s, z6.b, z0.b[3]\n"
      "sdot z14.s, z6.b, z1.b[3]\n"
      "sdot z18.s, z6.b, z2.b[3]\n"
      "sdot z11.s, z7.b, z0.b[3]\n"
      "sdot z15.s, z7.b, z1.b[3]\n"
      "sdot z19.s, z7.b, z2.b[3]\n"
      "35:"  // Height 3: Multiply loop: multiply skip
      "prfm pldl1keep, [x10, #0x80]\n"
      "add x12, x12, #0x1\n"
      "prfm pldl1keep, [x28, #0x80]\n"
      "prfm pldl1keep, [x26, #0x80]\n"
      "ldr w19, [%x[args_ptr], %[offsetof_num_strings]]\n"
      "cmp x12, x19\n"
      "bne 30b\n"
      "prfm pstl1keep, [x13, #0x0]\n"
      "prfm pstl1keep, [x9, #0x0]\n"
      "prfm pstl1keep, [x27, #0x0]\n"
      "st1w { z8.s }, p4, [x13]\n"
      "st1w { z9.s }, p3, [x13, #1, MUL VL]\n"
      "st1w { z10.s }, p2, [x13, #2, MUL VL]\n"
      "st1w { z11.s }, p1, [x13, #3, MUL VL]\n"
      "addvl x13, x13, #4\n"
      "st1w { z12.s }, p4, [x9]\n"
      "st1w { z13.s }, p3, [x9, #1, MUL VL]\n"
      "st1w { z14.s }, p2, [x9, #2, MUL VL]\n"
      "st1w { z15.s }, p1, [x9, #3, MUL VL]\n"
      "addvl x9, x9, #4\n"
      "st1w { z16.s }, p4, [x27]\n"
      "st1w { z17.s }, p3, [x27, #1, MUL VL]\n"
      "st1w { z18.s }, p2, [x27, #2, MUL VL]\n"
      "st1w { z19.s }, p1, [x27, #3, MUL VL]\n"
      "addvl x27, x27, #4\n"
      "36:"  // Height 3: Writeback done
      "mov x19, #0x0\n"
      "incw x19, ALL, MUL #4\n"
      "subs x15, x15, x19\n"
      "bgt 27b\n"
      "b 74f\n"
      "37:"  // Height 4
      "ldr x15, [%x[args_ptr], %[offsetof_N]]\n"
      "ldr x14, [%x[args_ptr], %[offsetof_B_ptr]]\n"
      "ldr x19, [%x[args_ptr], %[offsetof_output_offset]]\n"
      "tbz %x[flags], #2, 38f\n"
      "ldr x13, [%x[output_ptr], #0x0]\n"
      "add x13, x13, x19, LSL #2\n"
      "ldr x9, [%x[output_ptr], #0x8]\n"
      "ldr x27, [%x[output_ptr], #0x10]\n"
      "add x9, x9, x19, LSL #2\n"
      "ldr x25, [%x[output_ptr], #0x18]\n"
      "add x27, x27, x19, LSL #2\n"
      "add x25, x25, x19, LSL #2\n"
      "b 39f\n"
      "38:"  // Height 4: setup direct output
      "mov x13, %x[output_ptr]\n"
      "add x9, x13, x19, LSL #2\n"
      "add x27, x9, x19, LSL #2\n"
      "add x25, x27, x19, LSL #2\n"
      "39:"  // Height 4: Column loop
      "mov x19, #0x0\n"
      "whilelt p4.s, x19, x15\n"
      "incw x19\n"
      "whilelt p3.s, x19, x15\n"
      "incw x19\n"
      "whilelt p2.s, x19, x15\n"
      "incw x19\n"
      "whilelt p1.s, x19, x15\n"
      "tbz %x[flags], #0, 40f\n"
      "ld1w { z8.s }, p4/Z, [x13]\n"
      "ld1w { z9.s }, p3/Z, [x13, #1, MUL VL]\n"
      "ld1w { z10.s }, p2/Z, [x13, #2, MUL VL]\n"
      "ld1w { z11.s }, p1/Z, [x13, #3, MUL VL]\n"
      "ld1w { z12.s }, p4/Z, [x9]\n"
      "ld1w { z13.s }, p3/Z, [x9, #1, MUL VL]\n"
      "ld1w { z14.s }, p2/Z, [x9, #2, MUL VL]\n"
      "ld1w { z15.s }, p1/Z, [x9, #3, MUL VL]\n"
      "ld1w { z16.s }, p4/Z, [x27]\n"
      "ld1w { z17.s }, p3/Z, [x27, #1, MUL VL]\n"
      "ld1w { z18.s }, p2/Z, [x27, #2, MUL VL]\n"
      "ld1w { z19.s }, p1/Z, [x27, #3, MUL VL]\n"
      "ld1w { z20.s }, p4/Z, [x25]\n"
      "ld1w { z21.s }, p3/Z, [x25, #1, MUL VL]\n"
      "ld1w { z22.s }, p2/Z, [x25, #2, MUL VL]\n"
      "ld1w { z23.s }, p1/Z, [x25, #3, MUL VL]\n"
      "b 41f\n"
      "40:"  // Height 4: no accumulate
      "mov z8.s, #0x0\n"
      "mov z9.s, #0x0\n"
      "mov z10.s, #0x0\n"
      "mov z11.s, #0x0\n"
      "mov z12.s, #0x0\n"
      "mov z13.s, #0x0\n"
      "mov z14.s, #0x0\n"
      "mov z15.s, #0x0\n"
      "mov z16.s, #0x0\n"
      "mov z17.s, #0x0\n"
      "mov z18.s, #0x0\n"
      "mov z19.s, #0x0\n"
      "mov z20.s, #0x0\n"
      "mov z21.s, #0x0\n"
      "mov z22.s, #0x0\n"
      "mov z23.s, #0x0\n"
      "41:"  // Height 4: setup done
      "mov x12, #0x0\n"
      "42:"  // Height 4: String loop
      "ldr x20, [%x[args_ptr], %[offsetof_string_lengths]]\n"
      "ldr x19, [%x[args_ptr], %[offsetof_input_offset]]\n"
      "ldr w11, [x20, x12, LSL #0x2]\n"
      "tbz %x[flags], #3, 43f\n"
      "ldr x20, [%x[input_ptr], x12, LSL #0x3]\n"
      "add x20, x20, x19, LSL #3\n"
      "ldr x10, [x20, #0x0]\n"
      "ldr x28, [x20, #0x8]\n"
      "ldr x26, [x20, #0x10]\n"
      "ldr x24, [x20, #0x18]\n"
      "cbnz x12, 44f\n"
      "ldr x19, [%x[args_ptr], %[offsetof_input_initial_col]]\n"
      "add x10, x10, x19\n"
      "add x28, x28, x19\n"
      "add x26, x26, x19\n"
      "add x24, x24, x19\n"
      "b 44f\n"
      "43:"  // Height 4: setup direct input
      "mov x10, %x[input_ptr]\n"
      "add x28, x10, x19\n"
      "add x26, x28, x19\n"
      "add x24, x26, x19\n"
      "44:"  // Height 4: input setup done
      "cmp x11, #0x10\n"
      "ble 46f\n"
      "45:"  // Height 4: Multiply loop: Main loop head
      "ld1b { z6.b }, p5/Z, [x14]\n"
      "whilelt p0.b, XZR, x11\n"
      "ld1b { z7.b }, p5/Z, [x14, #1, MUL VL]\n"
      "sub x11, x11, #0x10\n"
      "ld1rqb { z0.b }, p0/Z, [x10]\n"
      "sdot z8.s, z6.b, z0.b[0]\n"
      "ld1rqb { z1.b }, p0/Z, [x28]\n"
      "add x10, x10, #0x10\n"
      "sdot z9.s, z7.b, z0.b[0]\n"
      "ld1rqb { z2.b }, p0/Z, [x26]\n"
      "add x28, x28, #0x10\n"
      "sdot z12.s, z6.b, z1.b[0]\n"
      "ld1rqb { z3.b }, p0/Z, [x24]\n"
      "add x26, x26, #0x10\n"
      "sdot z16.s, z6.b, z2.b[0]\n"
      "prfm pldl1keep, [x10, #0x80]\n"
      "add x24, x24, #0x10\n"
      "sdot z13.s, z7.b, z1.b[0]\n"
      "prfm pldl1keep, [x28, #0x80]\n"
      "cmp x11, #0x10\n"
      "sdot z20.s, z6.b, z3.b[0]\n"
      "ld1b { z6.b }, p5/Z, [x14, #2, MUL VL]\n"
      "sdot z17.s, z7.b, z2.b[0]\n"
      "prfm pldl1keep, [x26, #0x80]\n"
      "sdot z21.s, z7.b, z3.b[0]\n"
      "ld1b { z7.b }, p5/Z, [x14, #3, MUL VL]\n"
      "prfm pldl1keep, [x24, #0x80]\n"
      "sdot z10.s, z6.b, z0.b[0]\n"
      "sdot z14.s, z6.b, z1.b[0]\n"
      "sdot z18.s, z6.b, z2.b[0]\n"
      "sdot z22.s, z6.b, z3.b[0]\n"
      "ld1b { z6.b }, p5/Z, [x14, #4, MUL VL]\n"
      "sdot z11.s, z7.b, z0.b[0]\n"
      "sdot z15.s, z7.b, z1.b[0]\n"
      "sdot z19.s, z7.b, z2.b[0]\n"
      "sdot z23.s, z7.b, z3.b[0]\n"
      "ld1b { z7.b }, p5/Z, [x14, #5, MUL VL]\n"
      "sdot z8.s, z6.b, z0.b[1]\n"
      "sdot z12.s, z6.b, z1.b[1]\n"
      "sdot z16.s, z6.b, z2.b[1]\n"
      "sdot z20.s, z6.b, z3.b[1]\n"
      "ld1b { z6.b }, p5/Z, [x14, #6, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[1]\n"
      "sdot z13.s, z7.b, z1.b[1]\n"
      "sdot z17.s, z7.b, z2.b[1]\n"
      "sdot z21.s, z7.b, z3.b[1]\n"
      "ld1b { z7.b }, p5/Z, [x14, #7, MUL VL]\n"
      "addvl x14, x14, #16\n"
      "sdot z10.s, z6.b, z0.b[1]\n"
      "sdot z14.s, z6.b, z1.b[1]\n"
      "sdot z18.s, z6.b, z2.b[1]\n"
      "sdot z22.s, z6.b, z3.b[1]\n"
      "ld1b { z6.b }, p5/Z, [x14, #-8, MUL VL]\n"
      "sdot z11.s, z7.b, z0.b[1]\n"
      "sdot z15.s, z7.b, z1.b[1]\n"
      "sdot z19.s, z7.b, z2.b[1]\n"
      "sdot z23.s, z7.b, z3.b[1]\n"
      "ld1b { z7.b }, p5/Z, [x14, #-7, MUL VL]\n"
      "sdot z8.s, z6.b, z0.b[2]\n"
      "sdot z12.s, z6.b, z1.b[2]\n"
      "sdot z16.s, z6.b, z2.b[2]\n"
      "sdot z20.s, z6.b, z3.b[2]\n"
      "ld1b { z6.b }, p5/Z, [x14, #-6, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[2]\n"
      "sdot z13.s, z7.b, z1.b[2]\n"
      "sdot z17.s, z7.b, z2.b[2]\n"
      "sdot z21.s, z7.b, z3.b[2]\n"
      "ld1b { z7.b }, p5/Z, [x14, #-5, MUL VL]\n"
      "sdot z10.s, z6.b, z0.b[2]\n"
      "sdot z14.s, z6.b, z1.b[2]\n"
      "sdot z18.s, z6.b, z2.b[2]\n"
      "sdot z22.s, z6.b, z3.b[2]\n"
      "ld1b { z6.b }, p5/Z, [x14, #-4, MUL VL]\n"
      "sdot z11.s, z7.b, z0.b[2]\n"
      "sdot z15.s, z7.b, z1.b[2]\n"
      "sdot z19.s, z7.b, z2.b[2]\n"
      "sdot z23.s, z7.b, z3.b[2]\n"
      "ld1b { z7.b }, p5/Z, [x14, #-3, MUL VL]\n"
      "sdot z8.s, z6.b, z0.b[3]\n"
      "sdot z12.s, z6.b, z1.b[3]\n"
      "sdot z16.s, z6.b, z2.b[3]\n"
      "sdot z20.s, z6.b, z3.b[3]\n"
      "ld1b { z6.b }, p5/Z, [x14, #-2, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[3]\n"
      "sdot z13.s, z7.b, z1.b[3]\n"
      "sdot z17.s, z7.b, z2.b[3]\n"
      "sdot z21.s, z7.b, z3.b[3]\n"
      "ld1b { z7.b }, p5/Z, [x14, #-1, MUL VL]\n"
      "sdot z10.s, z6.b, z0.b[3]\n"
      "sdot z14.s, z6.b, z1.b[3]\n"
      "sdot z18.s, z6.b, z2.b[3]\n"
      "sdot z22.s, z6.b, z3.b[3]\n"
      "sdot z11.s, z7.b, z0.b[3]\n"
      "sdot z15.s, z7.b, z1.b[3]\n"
      "sdot z19.s, z7.b, z2.b[3]\n"
      "sdot z23.s, z7.b, z3.b[3]\n"
      "bgt 45b\n"
      "46:"  // Height 4: Multiply loop: Single iteration only
      "ld1b { z6.b }, p5/Z, [x14]\n"
      "whilelt p0.b, XZR, x11\n"
      "ld1b { z7.b }, p5/Z, [x14, #1, MUL VL]\n"
      "subs x11, x11, #0x4\n"
      "ld1rqb { z0.b }, p0/Z, [x10]\n"
      "sdot z8.s, z6.b, z0.b[0]\n"
      "ld1rqb { z1.b }, p0/Z, [x28]\n"
      "add x10, x10, #0x10\n"
      "sdot z9.s, z7.b, z0.b[0]\n"
      "ld1rqb { z2.b }, p0/Z, [x26]\n"
      "add x28, x28, #0x10\n"
      "sdot z12.s, z6.b, z1.b[0]\n"
      "ld1rqb { z3.b }, p0/Z, [x24]\n"
      "add x26, x26, #0x10\n"
      "sdot z16.s, z6.b, z2.b[0]\n"
      "add x24, x24, #0x10\n"
      "sdot z13.s, z7.b, z1.b[0]\n"
      "sdot z17.s, z7.b, z2.b[0]\n"
      "sdot z20.s, z6.b, z3.b[0]\n"
      "ld1b { z6.b }, p5/Z, [x14, #2, MUL VL]\n"
      "sdot z21.s, z7.b, z3.b[0]\n"
      "ld1b { z7.b }, p5/Z, [x14, #3, MUL VL]\n"
      "addvl x14, x14, #4\n"
      "sdot z10.s, z6.b, z0.b[0]\n"
      "sdot z14.s, z6.b, z1.b[0]\n"
      "sdot z18.s, z6.b, z2.b[0]\n"
      "sdot z22.s, z6.b, z3.b[0]\n"
      "sdot z11.s, z7.b, z0.b[0]\n"
      "sdot z15.s, z7.b, z1.b[0]\n"
      "sdot z19.s, z7.b, z2.b[0]\n"
      "sdot z23.s, z7.b, z3.b[0]\n"
      "ble 47f\n"
      "ld1b { z6.b }, p5/Z, [x14]\n"
      "sdot z8.s, z6.b, z0.b[1]\n"
      "ld1b { z7.b }, p5/Z, [x14, #1, MUL VL]\n"
      "subs x11, x11, #0x4\n"
      "sdot z12.s, z6.b, z1.b[1]\n"
      "sdot z16.s, z6.b, z2.b[1]\n"
      "sdot z20.s, z6.b, z3.b[1]\n"
      "ld1b { z6.b }, p5/Z, [x14, #2, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[1]\n"
      "sdot z13.s, z7.b, z1.b[1]\n"
      "sdot z17.s, z7.b, z2.b[1]\n"
      "sdot z21.s, z7.b, z3.b[1]\n"
      "ld1b { z7.b }, p5/Z, [x14, #3, MUL VL]\n"
      "addvl x14, x14, #4\n"
      "sdot z10.s, z6.b, z0.b[1]\n"
      "sdot z14.s, z6.b, z1.b[1]\n"
      "sdot z18.s, z6.b, z2.b[1]\n"
      "sdot z22.s, z6.b, z3.b[1]\n"
      "sdot z11.s, z7.b, z0.b[1]\n"
      "sdot z15.s, z7.b, z1.b[1]\n"
      "sdot z19.s, z7.b, z2.b[1]\n"
      "sdot z23.s, z7.b, z3.b[1]\n"
      "ble 47f\n"
      "ld1b { z6.b }, p5/Z, [x14]\n"
      "sdot z8.s, z6.b, z0.b[2]\n"
      "ld1b { z7.b }, p5/Z, [x14, #1, MUL VL]\n"
      "subs x11, x11, #0x4\n"
      "sdot z12.s, z6.b, z1.b[2]\n"
      "sdot z16.s, z6.b, z2.b[2]\n"
      "sdot z20.s, z6.b, z3.b[2]\n"
      "ld1b { z6.b }, p5/Z, [x14, #2, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[2]\n"
      "sdot z13.s, z7.b, z1.b[2]\n"
      "sdot z17.s, z7.b, z2.b[2]\n"
      "sdot z21.s, z7.b, z3.b[2]\n"
      "ld1b { z7.b }, p5/Z, [x14, #3, MUL VL]\n"
      "addvl x14, x14, #4\n"
      "sdot z10.s, z6.b, z0.b[2]\n"
      "sdot z14.s, z6.b, z1.b[2]\n"
      "sdot z18.s, z6.b, z2.b[2]\n"
      "sdot z22.s, z6.b, z3.b[2]\n"
      "sdot z11.s, z7.b, z0.b[2]\n"
      "sdot z15.s, z7.b, z1.b[2]\n"
      "sdot z19.s, z7.b, z2.b[2]\n"
      "sdot z23.s, z7.b, z3.b[2]\n"
      "ble 47f\n"
      "ld1b { z6.b }, p5/Z, [x14]\n"
      "sdot z8.s, z6.b, z0.b[3]\n"
      "ld1b { z7.b }, p5/Z, [x14, #1, MUL VL]\n"
      "sdot z12.s, z6.b, z1.b[3]\n"
      "sdot z16.s, z6.b, z2.b[3]\n"
      "sdot z20.s, z6.b, z3.b[3]\n"
      "ld1b { z6.b }, p5/Z, [x14, #2, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[3]\n"
      "sdot z13.s, z7.b, z1.b[3]\n"
      "sdot z17.s, z7.b, z2.b[3]\n"
      "sdot z21.s, z7.b, z3.b[3]\n"
      "ld1b { z7.b }, p5/Z, [x14, #3, MUL VL]\n"
      "addvl x14, x14, #4\n"
      "sdot z10.s, z6.b, z0.b[3]\n"
      "sdot z14.s, z6.b, z1.b[3]\n"
      "sdot z18.s, z6.b, z2.b[3]\n"
      "sdot z22.s, z6.b, z3.b[3]\n"
      "sdot z11.s, z7.b, z0.b[3]\n"
      "sdot z15.s, z7.b, z1.b[3]\n"
      "sdot z19.s, z7.b, z2.b[3]\n"
      "sdot z23.s, z7.b, z3.b[3]\n"
      "47:"  // Height 4: Multiply loop: multiply skip
      "prfm pldl1keep, [x10, #0x80]\n"
      "add x12, x12, #0x1\n"
      "prfm pldl1keep, [x28, #0x80]\n"
      "prfm pldl1keep, [x26, #0x80]\n"
      "prfm pldl1keep, [x24, #0x80]\n"
      "ldr w19, [%x[args_ptr], %[offsetof_num_strings]]\n"
      "cmp x12, x19\n"
      "bne 42b\n"
      "prfm pstl1keep, [x13, #0x0]\n"
      "prfm pstl1keep, [x9, #0x0]\n"
      "prfm pstl1keep, [x27, #0x0]\n"
      "prfm pstl1keep, [x25, #0x0]\n"
      "st1w { z8.s }, p4, [x13]\n"
      "st1w { z9.s }, p3, [x13, #1, MUL VL]\n"
      "st1w { z10.s }, p2, [x13, #2, MUL VL]\n"
      "st1w { z11.s }, p1, [x13, #3, MUL VL]\n"
      "addvl x13, x13, #4\n"
      "st1w { z12.s }, p4, [x9]\n"
      "st1w { z13.s }, p3, [x9, #1, MUL VL]\n"
      "st1w { z14.s }, p2, [x9, #2, MUL VL]\n"
      "st1w { z15.s }, p1, [x9, #3, MUL VL]\n"
      "addvl x9, x9, #4\n"
      "st1w { z16.s }, p4, [x27]\n"
      "st1w { z17.s }, p3, [x27, #1, MUL VL]\n"
      "st1w { z18.s }, p2, [x27, #2, MUL VL]\n"
      "st1w { z19.s }, p1, [x27, #3, MUL VL]\n"
      "addvl x27, x27, #4\n"
      "st1w { z20.s }, p4, [x25]\n"
      "st1w { z21.s }, p3, [x25, #1, MUL VL]\n"
      "st1w { z22.s }, p2, [x25, #2, MUL VL]\n"
      "st1w { z23.s }, p1, [x25, #3, MUL VL]\n"
      "addvl x25, x25, #4\n"
      "48:"  // Height 4: Writeback done
      "mov x19, #0x0\n"
      "incw x19, ALL, MUL #4\n"
      "subs x15, x15, x19\n"
      "bgt 39b\n"
      "b 74f\n"
      "49:"  // Height 5
      "ldr x15, [%x[args_ptr], %[offsetof_N]]\n"
      "ldr x14, [%x[args_ptr], %[offsetof_B_ptr]]\n"
      "ldr x19, [%x[args_ptr], %[offsetof_output_offset]]\n"
      "tbz %x[flags], #2, 50f\n"
      "ldr x13, [%x[output_ptr], #0x0]\n"
      "add x13, x13, x19, LSL #2\n"
      "ldr x9, [%x[output_ptr], #0x8]\n"
      "ldr x27, [%x[output_ptr], #0x10]\n"
      "add x9, x9, x19, LSL #2\n"
      "ldr x25, [%x[output_ptr], #0x18]\n"
      "ldr x23, [%x[output_ptr], #0x20]\n"
      "add x27, x27, x19, LSL #2\n"
      "add x25, x25, x19, LSL #2\n"
      "add x23, x23, x19, LSL #2\n"
      "b 51f\n"
      "50:"  // Height 5: setup direct output
      "mov x13, %x[output_ptr]\n"
      "add x9, x13, x19, LSL #2\n"
      "add x27, x9, x19, LSL #2\n"
      "add x25, x27, x19, LSL #2\n"
      "add x23, x25, x19, LSL #2\n"
      "51:"  // Height 5: Column loop
      "mov x19, #0x0\n"
      "whilelt p4.s, x19, x15\n"
      "incw x19\n"
      "whilelt p3.s, x19, x15\n"
      "incw x19\n"
      "whilelt p2.s, x19, x15\n"
      "incw x19\n"
      "whilelt p1.s, x19, x15\n"
      "tbz %x[flags], #0, 52f\n"
      "ld1w { z8.s }, p4/Z, [x13]\n"
      "ld1w { z9.s }, p3/Z, [x13, #1, MUL VL]\n"
      "ld1w { z10.s }, p2/Z, [x13, #2, MUL VL]\n"
      "ld1w { z11.s }, p1/Z, [x13, #3, MUL VL]\n"
      "ld1w { z12.s }, p4/Z, [x9]\n"
      "ld1w { z13.s }, p3/Z, [x9, #1, MUL VL]\n"
      "ld1w { z14.s }, p2/Z, [x9, #2, MUL VL]\n"
      "ld1w { z15.s }, p1/Z, [x9, #3, MUL VL]\n"
      "ld1w { z16.s }, p4/Z, [x27]\n"
      "ld1w { z17.s }, p3/Z, [x27, #1, MUL VL]\n"
      "ld1w { z18.s }, p2/Z, [x27, #2, MUL VL]\n"
      "ld1w { z19.s }, p1/Z, [x27, #3, MUL VL]\n"
      "ld1w { z20.s }, p4/Z, [x25]\n"
      "ld1w { z21.s }, p3/Z, [x25, #1, MUL VL]\n"
      "ld1w { z22.s }, p2/Z, [x25, #2, MUL VL]\n"
      "ld1w { z23.s }, p1/Z, [x25, #3, MUL VL]\n"
      "ld1w { z24.s }, p4/Z, [x23]\n"
      "ld1w { z25.s }, p3/Z, [x23, #1, MUL VL]\n"
      "ld1w { z26.s }, p2/Z, [x23, #2, MUL VL]\n"
      "ld1w { z27.s }, p1/Z, [x23, #3, MUL VL]\n"
      "b 53f\n"
      "52:"  // Height 5: no accumulate
      "mov z8.s, #0x0\n"
      "mov z9.s, #0x0\n"
      "mov z10.s, #0x0\n"
      "mov z11.s, #0x0\n"
      "mov z12.s, #0x0\n"
      "mov z13.s, #0x0\n"
      "mov z14.s, #0x0\n"
      "mov z15.s, #0x0\n"
      "mov z16.s, #0x0\n"
      "mov z17.s, #0x0\n"
      "mov z18.s, #0x0\n"
      "mov z19.s, #0x0\n"
      "mov z20.s, #0x0\n"
      "mov z21.s, #0x0\n"
      "mov z22.s, #0x0\n"
      "mov z23.s, #0x0\n"
      "mov z24.s, #0x0\n"
      "mov z25.s, #0x0\n"
      "mov z26.s, #0x0\n"
      "mov z27.s, #0x0\n"
      "53:"  // Height 5: setup done
      "mov x12, #0x0\n"
      "54:"  // Height 5: String loop
      "ldr x20, [%x[args_ptr], %[offsetof_string_lengths]]\n"
      "ldr x19, [%x[args_ptr], %[offsetof_input_offset]]\n"
      "ldr w11, [x20, x12, LSL #0x2]\n"
      "tbz %x[flags], #3, 55f\n"
      "ldr x20, [%x[input_ptr], x12, LSL #0x3]\n"
      "add x20, x20, x19, LSL #3\n"
      "ldr x10, [x20, #0x0]\n"
      "ldr x28, [x20, #0x8]\n"
      "ldr x26, [x20, #0x10]\n"
      "ldr x24, [x20, #0x18]\n"
      "ldr x22, [x20, #0x20]\n"
      "cbnz x12, 56f\n"
      "ldr x19, [%x[args_ptr], %[offsetof_input_initial_col]]\n"
      "add x10, x10, x19\n"
      "add x28, x28, x19\n"
      "add x26, x26, x19\n"
      "add x24, x24, x19\n"
      "add x22, x22, x19\n"
      "b 56f\n"
      "55:"  // Height 5: setup direct input
      "mov x10, %x[input_ptr]\n"
      "add x28, x10, x19\n"
      "add x26, x28, x19\n"
      "add x24, x26, x19\n"
      "add x22, x24, x19\n"
      "56:"  // Height 5: input setup done
      "cmp x11, #0x10\n"
      "ble 58f\n"
      "57:"  // Height 5: Multiply loop: Main loop head
      "ld1b { z6.b }, p5/Z, [x14]\n"
      "whilelt p0.b, XZR, x11\n"
      "ld1b { z7.b }, p5/Z, [x14, #1, MUL VL]\n"
      "sub x11, x11, #0x10\n"
      "ld1rqb { z0.b }, p0/Z, [x10]\n"
      "sdot z8.s, z6.b, z0.b[0]\n"
      "ld1rqb { z1.b }, p0/Z, [x28]\n"
      "add x10, x10, #0x10\n"
      "sdot z9.s, z7.b, z0.b[0]\n"
      "ld1rqb { z2.b }, p0/Z, [x26]\n"
      "add x28, x28, #0x10\n"
      "sdot z12.s, z6.b, z1.b[0]\n"
      "ld1rqb { z3.b }, p0/Z, [x24]\n"
      "add x26, x26, #0x10\n"
      "sdot z16.s, z6.b, z2.b[0]\n"
      "ld1rqb { z4.b }, p0/Z, [x22]\n"
      "add x24, x24, #0x10\n"
      "sdot z13.s, z7.b, z1.b[0]\n"
      "prfm pldl1keep, [x10, #0x80]\n"
      "add x22, x22, #0x10\n"
      "sdot z20.s, z6.b, z3.b[0]\n"
      "prfm pldl1keep, [x28, #0x80]\n"
      "cmp x11, #0x10\n"
      "sdot z24.s, z6.b, z4.b[0]\n"
      "ld1b { z6.b }, p5/Z, [x14, #2, MUL VL]\n"
      "sdot z17.s, z7.b, z2.b[0]\n"
      "prfm pldl1keep, [x26, #0x80]\n"
      "sdot z21.s, z7.b, z3.b[0]\n"
      "prfm pldl1keep, [x24, #0x80]\n"
      "sdot z25.s, z7.b, z4.b[0]\n"
      "ld1b { z7.b }, p5/Z, [x14, #3, MUL VL]\n"
      "sdot z10.s, z6.b, z0.b[0]\n"
      "prfm pldl1keep, [x22, #0x80]\n"
      "sdot z14.s, z6.b, z1.b[0]\n"
      "sdot z18.s, z6.b, z2.b[0]\n"
      "sdot z22.s, z6.b, z3.b[0]\n"
      "sdot z26.s, z6.b, z4.b[0]\n"
      "ld1b { z6.b }, p5/Z, [x14, #4, MUL VL]\n"
      "sdot z11.s, z7.b, z0.b[0]\n"
      "sdot z15.s, z7.b, z1.b[0]\n"
      "sdot z19.s, z7.b, z2.b[0]\n"
      "sdot z23.s, z7.b, z3.b[0]\n"
      "sdot z27.s, z7.b, z4.b[0]\n"
      "ld1b { z7.b }, p5/Z, [x14, #5, MUL VL]\n"
      "sdot z8.s, z6.b, z0.b[1]\n"
      "sdot z12.s, z6.b, z1.b[1]\n"
      "sdot z16.s, z6.b, z2.b[1]\n"
      "sdot z20.s, z6.b, z3.b[1]\n"
      "sdot z24.s, z6.b, z4.b[1]\n"
      "ld1b { z6.b }, p5/Z, [x14, #6, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[1]\n"
      "sdot z13.s, z7.b, z1.b[1]\n"
      "sdot z17.s, z7.b, z2.b[1]\n"
      "sdot z21.s, z7.b, z3.b[1]\n"
      "sdot z25.s, z7.b, z4.b[1]\n"
      "ld1b { z7.b }, p5/Z, [x14, #7, MUL VL]\n"
      "addvl x14, x14, #16\n"
      "sdot z10.s, z6.b, z0.b[1]\n"
      "sdot z14.s, z6.b, z1.b[1]\n"
      "sdot z18.s, z6.b, z2.b[1]\n"
      "sdot z22.s, z6.b, z3.b[1]\n"
      "sdot z26.s, z6.b, z4.b[1]\n"
      "ld1b { z6.b }, p5/Z, [x14, #-8, MUL VL]\n"
      "sdot z11.s, z7.b, z0.b[1]\n"
      "sdot z15.s, z7.b, z1.b[1]\n"
      "sdot z19.s, z7.b, z2.b[1]\n"
      "sdot z23.s, z7.b, z3.b[1]\n"
      "sdot z27.s, z7.b, z4.b[1]\n"
      "ld1b { z7.b }, p5/Z, [x14, #-7, MUL VL]\n"
      "sdot z8.s, z6.b, z0.b[2]\n"
      "sdot z12.s, z6.b, z1.b[2]\n"
      "sdot z16.s, z6.b, z2.b[2]\n"
      "sdot z20.s, z6.b, z3.b[2]\n"
      "sdot z24.s, z6.b, z4.b[2]\n"
      "ld1b { z6.b }, p5/Z, [x14, #-6, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[2]\n"
      "sdot z13.s, z7.b, z1.b[2]\n"
      "sdot z17.s, z7.b, z2.b[2]\n"
      "sdot z21.s, z7.b, z3.b[2]\n"
      "sdot z25.s, z7.b, z4.b[2]\n"
      "ld1b { z7.b }, p5/Z, [x14, #-5, MUL VL]\n"
      "sdot z10.s, z6.b, z0.b[2]\n"
      "sdot z14.s, z6.b, z1.b[2]\n"
      "sdot z18.s, z6.b, z2.b[2]\n"
      "sdot z22.s, z6.b, z3.b[2]\n"
      "sdot z26.s, z6.b, z4.b[2]\n"
      "ld1b { z6.b }, p5/Z, [x14, #-4, MUL VL]\n"
      "sdot z11.s, z7.b, z0.b[2]\n"
      "sdot z15.s, z7.b, z1.b[2]\n"
      "sdot z19.s, z7.b, z2.b[2]\n"
      "sdot z23.s, z7.b, z3.b[2]\n"
      "sdot z27.s, z7.b, z4.b[2]\n"
      "ld1b { z7.b }, p5/Z, [x14, #-3, MUL VL]\n"
      "sdot z8.s, z6.b, z0.b[3]\n"
      "sdot z12.s, z6.b, z1.b[3]\n"
      "sdot z16.s, z6.b, z2.b[3]\n"
      "sdot z20.s, z6.b, z3.b[3]\n"
      "sdot z24.s, z6.b, z4.b[3]\n"
      "ld1b { z6.b }, p5/Z, [x14, #-2, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[3]\n"
      "sdot z13.s, z7.b, z1.b[3]\n"
      "sdot z17.s, z7.b, z2.b[3]\n"
      "sdot z21.s, z7.b, z3.b[3]\n"
      "sdot z25.s, z7.b, z4.b[3]\n"
      "ld1b { z7.b }, p5/Z, [x14, #-1, MUL VL]\n"
      "sdot z10.s, z6.b, z0.b[3]\n"
      "sdot z14.s, z6.b, z1.b[3]\n"
      "sdot z18.s, z6.b, z2.b[3]\n"
      "sdot z22.s, z6.b, z3.b[3]\n"
      "sdot z26.s, z6.b, z4.b[3]\n"
      "sdot z11.s, z7.b, z0.b[3]\n"
      "sdot z15.s, z7.b, z1.b[3]\n"
      "sdot z19.s, z7.b, z2.b[3]\n"
      "sdot z23.s, z7.b, z3.b[3]\n"
      "sdot z27.s, z7.b, z4.b[3]\n"
      "bgt 57b\n"
      "58:"  // Height 5: Multiply loop: Single iteration only
      "ld1b { z6.b }, p5/Z, [x14]\n"
      "whilelt p0.b, XZR, x11\n"
      "ld1b { z7.b }, p5/Z, [x14, #1, MUL VL]\n"
      "subs x11, x11, #0x4\n"
      "ld1rqb { z0.b }, p0/Z, [x10]\n"
      "sdot z8.s, z6.b, z0.b[0]\n"
      "ld1rqb { z1.b }, p0/Z, [x28]\n"
      "add x10, x10, #0x10\n"
      "sdot z9.s, z7.b, z0.b[0]\n"
      "ld1rqb { z2.b }, p0/Z, [x26]\n"
      "add x28, x28, #0x10\n"
      "sdot z12.s, z6.b, z1.b[0]\n"
      "ld1rqb { z3.b }, p0/Z, [x24]\n"
      "add x26, x26, #0x10\n"
      "sdot z16.s, z6.b, z2.b[0]\n"
      "ld1rqb { z4.b }, p0/Z, [x22]\n"
      "add x24, x24, #0x10\n"
      "sdot z13.s, z7.b, z1.b[0]\n"
      "add x22, x22, #0x10\n"
      "sdot z17.s, z7.b, z2.b[0]\n"
      "sdot z20.s, z6.b, z3.b[0]\n"
      "sdot z24.s, z6.b, z4.b[0]\n"
      "ld1b { z6.b }, p5/Z, [x14, #2, MUL VL]\n"
      "sdot z21.s, z7.b, z3.b[0]\n"
      "sdot z25.s, z7.b, z4.b[0]\n"
      "ld1b { z7.b }, p5/Z, [x14, #3, MUL VL]\n"
      "addvl x14, x14, #4\n"
      "sdot z10.s, z6.b, z0.b[0]\n"
      "sdot z14.s, z6.b, z1.b[0]\n"
      "sdot z18.s, z6.b, z2.b[0]\n"
      "sdot z22.s, z6.b, z3.b[0]\n"
      "sdot z26.s, z6.b, z4.b[0]\n"
      "sdot z11.s, z7.b, z0.b[0]\n"
      "sdot z15.s, z7.b, z1.b[0]\n"
      "sdot z19.s, z7.b, z2.b[0]\n"
      "sdot z23.s, z7.b, z3.b[0]\n"
      "sdot z27.s, z7.b, z4.b[0]\n"
      "ble 59f\n"
      "ld1b { z6.b }, p5/Z, [x14]\n"
      "sdot z8.s, z6.b, z0.b[1]\n"
      "ld1b { z7.b }, p5/Z, [x14, #1, MUL VL]\n"
      "subs x11, x11, #0x4\n"
      "sdot z12.s, z6.b, z1.b[1]\n"
      "sdot z16.s, z6.b, z2.b[1]\n"
      "sdot z20.s, z6.b, z3.b[1]\n"
      "sdot z24.s, z6.b, z4.b[1]\n"
      "ld1b { z6.b }, p5/Z, [x14, #2, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[1]\n"
      "sdot z13.s, z7.b, z1.b[1]\n"
      "sdot z17.s, z7.b, z2.b[1]\n"
      "sdot z21.s, z7.b, z3.b[1]\n"
      "sdot z25.s, z7.b, z4.b[1]\n"
      "ld1b { z7.b }, p5/Z, [x14, #3, MUL VL]\n"
      "addvl x14, x14, #4\n"
      "sdot z10.s, z6.b, z0.b[1]\n"
      "sdot z14.s, z6.b, z1.b[1]\n"
      "sdot z18.s, z6.b, z2.b[1]\n"
      "sdot z22.s, z6.b, z3.b[1]\n"
      "sdot z26.s, z6.b, z4.b[1]\n"
      "sdot z11.s, z7.b, z0.b[1]\n"
      "sdot z15.s, z7.b, z1.b[1]\n"
      "sdot z19.s, z7.b, z2.b[1]\n"
      "sdot z23.s, z7.b, z3.b[1]\n"
      "sdot z27.s, z7.b, z4.b[1]\n"
      "ble 59f\n"
      "ld1b { z6.b }, p5/Z, [x14]\n"
      "sdot z8.s, z6.b, z0.b[2]\n"
      "ld1b { z7.b }, p5/Z, [x14, #1, MUL VL]\n"
      "subs x11, x11, #0x4\n"
      "sdot z12.s, z6.b, z1.b[2]\n"
      "sdot z16.s, z6.b, z2.b[2]\n"
      "sdot z20.s, z6.b, z3.b[2]\n"
      "sdot z24.s, z6.b, z4.b[2]\n"
      "ld1b { z6.b }, p5/Z, [x14, #2, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[2]\n"
      "sdot z13.s, z7.b, z1.b[2]\n"
      "sdot z17.s, z7.b, z2.b[2]\n"
      "sdot z21.s, z7.b, z3.b[2]\n"
      "sdot z25.s, z7.b, z4.b[2]\n"
      "ld1b { z7.b }, p5/Z, [x14, #3, MUL VL]\n"
      "addvl x14, x14, #4\n"
      "sdot z10.s, z6.b, z0.b[2]\n"
      "sdot z14.s, z6.b, z1.b[2]\n"
      "sdot z18.s, z6.b, z2.b[2]\n"
      "sdot z22.s, z6.b, z3.b[2]\n"
      "sdot z26.s, z6.b, z4.b[2]\n"
      "sdot z11.s, z7.b, z0.b[2]\n"
      "sdot z15.s, z7.b, z1.b[2]\n"
      "sdot z19.s, z7.b, z2.b[2]\n"
      "sdot z23.s, z7.b, z3.b[2]\n"
      "sdot z27.s, z7.b, z4.b[2]\n"
      "ble 59f\n"
      "ld1b { z6.b }, p5/Z, [x14]\n"
      "sdot z8.s, z6.b, z0.b[3]\n"
      "ld1b { z7.b }, p5/Z, [x14, #1, MUL VL]\n"
      "sdot z12.s, z6.b, z1.b[3]\n"
      "sdot z16.s, z6.b, z2.b[3]\n"
      "sdot z20.s, z6.b, z3.b[3]\n"
      "sdot z24.s, z6.b, z4.b[3]\n"
      "ld1b { z6.b }, p5/Z, [x14, #2, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[3]\n"
      "sdot z13.s, z7.b, z1.b[3]\n"
      "sdot z17.s, z7.b, z2.b[3]\n"
      "sdot z21.s, z7.b, z3.b[3]\n"
      "sdot z25.s, z7.b, z4.b[3]\n"
      "ld1b { z7.b }, p5/Z, [x14, #3, MUL VL]\n"
      "addvl x14, x14, #4\n"
      "sdot z10.s, z6.b, z0.b[3]\n"
      "sdot z14.s, z6.b, z1.b[3]\n"
      "sdot z18.s, z6.b, z2.b[3]\n"
      "sdot z22.s, z6.b, z3.b[3]\n"
      "sdot z26.s, z6.b, z4.b[3]\n"
      "sdot z11.s, z7.b, z0.b[3]\n"
      "sdot z15.s, z7.b, z1.b[3]\n"
      "sdot z19.s, z7.b, z2.b[3]\n"
      "sdot z23.s, z7.b, z3.b[3]\n"
      "sdot z27.s, z7.b, z4.b[3]\n"
      "59:"  // Height 5: Multiply loop: multiply skip
      "prfm pldl1keep, [x10, #0x80]\n"
      "add x12, x12, #0x1\n"
      "prfm pldl1keep, [x28, #0x80]\n"
      "prfm pldl1keep, [x26, #0x80]\n"
      "prfm pldl1keep, [x24, #0x80]\n"
      "prfm pldl1keep, [x22, #0x80]\n"
      "ldr w19, [%x[args_ptr], %[offsetof_num_strings]]\n"
      "cmp x12, x19\n"
      "bne 54b\n"
      "prfm pstl1keep, [x13, #0x0]\n"
      "prfm pstl1keep, [x9, #0x0]\n"
      "prfm pstl1keep, [x27, #0x0]\n"
      "prfm pstl1keep, [x25, #0x0]\n"
      "prfm pstl1keep, [x23, #0x0]\n"
      "st1w { z8.s }, p4, [x13]\n"
      "st1w { z9.s }, p3, [x13, #1, MUL VL]\n"
      "st1w { z10.s }, p2, [x13, #2, MUL VL]\n"
      "st1w { z11.s }, p1, [x13, #3, MUL VL]\n"
      "addvl x13, x13, #4\n"
      "st1w { z12.s }, p4, [x9]\n"
      "st1w { z13.s }, p3, [x9, #1, MUL VL]\n"
      "st1w { z14.s }, p2, [x9, #2, MUL VL]\n"
      "st1w { z15.s }, p1, [x9, #3, MUL VL]\n"
      "addvl x9, x9, #4\n"
      "st1w { z16.s }, p4, [x27]\n"
      "st1w { z17.s }, p3, [x27, #1, MUL VL]\n"
      "st1w { z18.s }, p2, [x27, #2, MUL VL]\n"
      "st1w { z19.s }, p1, [x27, #3, MUL VL]\n"
      "addvl x27, x27, #4\n"
      "st1w { z20.s }, p4, [x25]\n"
      "st1w { z21.s }, p3, [x25, #1, MUL VL]\n"
      "st1w { z22.s }, p2, [x25, #2, MUL VL]\n"
      "st1w { z23.s }, p1, [x25, #3, MUL VL]\n"
      "addvl x25, x25, #4\n"
      "st1w { z24.s }, p4, [x23]\n"
      "st1w { z25.s }, p3, [x23, #1, MUL VL]\n"
      "st1w { z26.s }, p2, [x23, #2, MUL VL]\n"
      "st1w { z27.s }, p1, [x23, #3, MUL VL]\n"
      "addvl x23, x23, #4\n"
      "60:"  // Height 5: Writeback done
      "mov x19, #0x0\n"
      "incw x19, ALL, MUL #4\n"
      "subs x15, x15, x19\n"
      "bgt 51b\n"
      "b 74f\n"
      "61:"  // Height 6
      "ldr x15, [%x[args_ptr], %[offsetof_N]]\n"
      "ldr x14, [%x[args_ptr], %[offsetof_B_ptr]]\n"
      "ldr x19, [%x[args_ptr], %[offsetof_output_offset]]\n"
      "tbz %x[flags], #2, 62f\n"
      "ldr x13, [%x[output_ptr], #0x0]\n"
      "add x13, x13, x19, LSL #2\n"
      "ldr x9, [%x[output_ptr], #0x8]\n"
      "ldr x27, [%x[output_ptr], #0x10]\n"
      "add x9, x9, x19, LSL #2\n"
      "ldr x25, [%x[output_ptr], #0x18]\n"
      "ldr x23, [%x[output_ptr], #0x20]\n"
      "add x27, x27, x19, LSL #2\n"
      "ldr x21, [%x[output_ptr], #0x28]\n"
      "add %x[output_ptr], %x[output_ptr], #0x30\n"
      "add x25, x25, x19, LSL #2\n"
      "add x23, x23, x19, LSL #2\n"
      "add x21, x21, x19, LSL #2\n"
      "b 63f\n"
      "62:"  // Height 6: setup direct output
      "mov x13, %x[output_ptr]\n"
      "add x9, x13, x19, LSL #2\n"
      "add x27, x9, x19, LSL #2\n"
      "add x25, x27, x19, LSL #2\n"
      "add x23, x25, x19, LSL #2\n"
      "add x21, x23, x19, LSL #2\n"
      "add %x[output_ptr], x21, x19, LSL #2\n"
      "63:"  // Height 6: Column loop
      "mov x19, #0x0\n"
      "whilelt p4.s, x19, x15\n"
      "incw x19\n"
      "whilelt p3.s, x19, x15\n"
      "incw x19\n"
      "whilelt p2.s, x19, x15\n"
      "incw x19\n"
      "whilelt p1.s, x19, x15\n"
      "tbz %x[flags], #0, 64f\n"
      "ld1w { z8.s }, p4/Z, [x13]\n"
      "ld1w { z9.s }, p3/Z, [x13, #1, MUL VL]\n"
      "ld1w { z10.s }, p2/Z, [x13, #2, MUL VL]\n"
      "ld1w { z11.s }, p1/Z, [x13, #3, MUL VL]\n"
      "ld1w { z12.s }, p4/Z, [x9]\n"
      "ld1w { z13.s }, p3/Z, [x9, #1, MUL VL]\n"
      "ld1w { z14.s }, p2/Z, [x9, #2, MUL VL]\n"
      "ld1w { z15.s }, p1/Z, [x9, #3, MUL VL]\n"
      "ld1w { z16.s }, p4/Z, [x27]\n"
      "ld1w { z17.s }, p3/Z, [x27, #1, MUL VL]\n"
      "ld1w { z18.s }, p2/Z, [x27, #2, MUL VL]\n"
      "ld1w { z19.s }, p1/Z, [x27, #3, MUL VL]\n"
      "ld1w { z20.s }, p4/Z, [x25]\n"
      "ld1w { z21.s }, p3/Z, [x25, #1, MUL VL]\n"
      "ld1w { z22.s }, p2/Z, [x25, #2, MUL VL]\n"
      "ld1w { z23.s }, p1/Z, [x25, #3, MUL VL]\n"
      "ld1w { z24.s }, p4/Z, [x23]\n"
      "ld1w { z25.s }, p3/Z, [x23, #1, MUL VL]\n"
      "ld1w { z26.s }, p2/Z, [x23, #2, MUL VL]\n"
      "ld1w { z27.s }, p1/Z, [x23, #3, MUL VL]\n"
      "ld1w { z28.s }, p4/Z, [x21]\n"
      "ld1w { z29.s }, p3/Z, [x21, #1, MUL VL]\n"
      "ld1w { z30.s }, p2/Z, [x21, #2, MUL VL]\n"
      "ld1w { z31.s }, p1/Z, [x21, #3, MUL VL]\n"
      "b 65f\n"
      "64:"  // Height 6: no accumulate
      "mov z8.s, #0x0\n"
      "mov z9.s, #0x0\n"
      "mov z10.s, #0x0\n"
      "mov z11.s, #0x0\n"
      "mov z12.s, #0x0\n"
      "mov z13.s, #0x0\n"
      "mov z14.s, #0x0\n"
      "mov z15.s, #0x0\n"
      "mov z16.s, #0x0\n"
      "mov z17.s, #0x0\n"
      "mov z18.s, #0x0\n"
      "mov z19.s, #0x0\n"
      "mov z20.s, #0x0\n"
      "mov z21.s, #0x0\n"
      "mov z22.s, #0x0\n"
      "mov z23.s, #0x0\n"
      "mov z24.s, #0x0\n"
      "mov z25.s, #0x0\n"
      "mov z26.s, #0x0\n"
      "mov z27.s, #0x0\n"
      "mov z28.s, #0x0\n"
      "mov z29.s, #0x0\n"
      "mov z30.s, #0x0\n"
      "mov z31.s, #0x0\n"
      "65:"  // Height 6: setup done
      "mov x12, #0x0\n"
      "66:"  // Height 6: String loop
      "ldr x20, [%x[args_ptr], %[offsetof_string_lengths]]\n"
      "ldr x19, [%x[args_ptr], %[offsetof_input_offset]]\n"
      "ldr w11, [x20, x12, LSL #0x2]\n"
      "tbz %x[flags], #3, 67f\n"
      "ldr x20, [%x[input_ptr], x12, LSL #0x3]\n"
      "add x20, x20, x19, LSL #3\n"
      "ldr x10, [x20, #0x0]\n"
      "ldr x28, [x20, #0x8]\n"
      "ldr x26, [x20, #0x10]\n"
      "ldr x24, [x20, #0x18]\n"
      "ldr x22, [x20, #0x20]\n"
      "ldr x20, [x20, #0x28]\n"
      "cbnz x12, 68f\n"
      "ldr x19, [%x[args_ptr], %[offsetof_input_initial_col]]\n"
      "add x10, x10, x19\n"
      "add x28, x28, x19\n"
      "add x26, x26, x19\n"
      "add x24, x24, x19\n"
      "add x22, x22, x19\n"
      "add x20, x20, x19\n"
      "b 68f\n"
      "67:"  // Height 6: setup direct input
      "mov x10, %x[input_ptr]\n"
      "add x28, x10, x19\n"
      "add x26, x28, x19\n"
      "add x24, x26, x19\n"
      "add x22, x24, x19\n"
      "add x20, x22, x19\n"
      "68:"  // Height 6: input setup done
      "cmp x11, #0x10\n"
      "ble 70f\n"
      "69:"  // Height 6: Multiply loop: Main loop head
      "ld1b { z6.b }, p5/Z, [x14]\n"
      "whilelt p0.b, XZR, x11\n"
      "ld1b { z7.b }, p5/Z, [x14, #1, MUL VL]\n"
      "sub x11, x11, #0x10\n"
      "ld1rqb { z0.b }, p0/Z, [x10]\n"
      "sdot z8.s, z6.b, z0.b[0]\n"
      "ld1rqb { z1.b }, p0/Z, [x28]\n"
      "add x10, x10, #0x10\n"
      "sdot z9.s, z7.b, z0.b[0]\n"
      "ld1rqb { z2.b }, p0/Z, [x26]\n"
      "add x28, x28, #0x10\n"
      "sdot z12.s, z6.b, z1.b[0]\n"
      "ld1rqb { z3.b }, p0/Z, [x24]\n"
      "add x26, x26, #0x10\n"
      "sdot z16.s, z6.b, z2.b[0]\n"
      "ld1rqb { z4.b }, p0/Z, [x22]\n"
      "add x24, x24, #0x10\n"
      "sdot z13.s, z7.b, z1.b[0]\n"
      "ld1rqb { z5.b }, p0/Z, [x20]\n"
      "add x22, x22, #0x10\n"
      "sdot z20.s, z6.b, z3.b[0]\n"
      "prfm pldl1keep, [x10, #0x80]\n"
      "add x20, x20, #0x10\n"
      "sdot z24.s, z6.b, z4.b[0]\n"
      "prfm pldl1keep, [x28, #0x80]\n"
      "cmp x11, #0x10\n"
      "sdot z28.s, z6.b, z5.b[0]\n"
      "ld1b { z6.b }, p5/Z, [x14, #2, MUL VL]\n"
      "sdot z17.s, z7.b, z2.b[0]\n"
      "prfm pldl1keep, [x26, #0x80]\n"
      "sdot z21.s, z7.b, z3.b[0]\n"
      "prfm pldl1keep, [x24, #0x80]\n"
      "sdot z25.s, z7.b, z4.b[0]\n"
      "prfm pldl1keep, [x22, #0x80]\n"
      "sdot z29.s, z7.b, z5.b[0]\n"
      "ld1b { z7.b }, p5/Z, [x14, #3, MUL VL]\n"
      "sdot z10.s, z6.b, z0.b[0]\n"
      "prfm pldl1keep, [x20, #0x80]\n"
      "sdot z14.s, z6.b, z1.b[0]\n"
      "sdot z18.s, z6.b, z2.b[0]\n"
      "sdot z22.s, z6.b, z3.b[0]\n"
      "sdot z26.s, z6.b, z4.b[0]\n"
      "sdot z30.s, z6.b, z5.b[0]\n"
      "ld1b { z6.b }, p5/Z, [x14, #4, MUL VL]\n"
      "sdot z11.s, z7.b, z0.b[0]\n"
      "sdot z15.s, z7.b, z1.b[0]\n"
      "sdot z19.s, z7.b, z2.b[0]\n"
      "sdot z23.s, z7.b, z3.b[0]\n"
      "sdot z27.s, z7.b, z4.b[0]\n"
      "sdot z31.s, z7.b, z5.b[0]\n"
      "ld1b { z7.b }, p5/Z, [x14, #5, MUL VL]\n"
      "sdot z8.s, z6.b, z0.b[1]\n"
      "sdot z12.s, z6.b, z1.b[1]\n"
      "sdot z16.s, z6.b, z2.b[1]\n"
      "sdot z20.s, z6.b, z3.b[1]\n"
      "sdot z24.s, z6.b, z4.b[1]\n"
      "sdot z28.s, z6.b, z5.b[1]\n"
      "ld1b { z6.b }, p5/Z, [x14, #6, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[1]\n"
      "sdot z13.s, z7.b, z1.b[1]\n"
      "sdot z17.s, z7.b, z2.b[1]\n"
      "sdot z21.s, z7.b, z3.b[1]\n"
      "sdot z25.s, z7.b, z4.b[1]\n"
      "sdot z29.s, z7.b, z5.b[1]\n"
      "ld1b { z7.b }, p5/Z, [x14, #7, MUL VL]\n"
      "addvl x14, x14, #16\n"
      "sdot z10.s, z6.b, z0.b[1]\n"
      "sdot z14.s, z6.b, z1.b[1]\n"
      "sdot z18.s, z6.b, z2.b[1]\n"
      "sdot z22.s, z6.b, z3.b[1]\n"
      "sdot z26.s, z6.b, z4.b[1]\n"
      "sdot z30.s, z6.b, z5.b[1]\n"
      "ld1b { z6.b }, p5/Z, [x14, #-8, MUL VL]\n"
      "sdot z11.s, z7.b, z0.b[1]\n"
      "sdot z15.s, z7.b, z1.b[1]\n"
      "sdot z19.s, z7.b, z2.b[1]\n"
      "sdot z23.s, z7.b, z3.b[1]\n"
      "sdot z27.s, z7.b, z4.b[1]\n"
      "sdot z31.s, z7.b, z5.b[1]\n"
      "ld1b { z7.b }, p5/Z, [x14, #-7, MUL VL]\n"
      "sdot z8.s, z6.b, z0.b[2]\n"
      "sdot z12.s, z6.b, z1.b[2]\n"
      "sdot z16.s, z6.b, z2.b[2]\n"
      "sdot z20.s, z6.b, z3.b[2]\n"
      "sdot z24.s, z6.b, z4.b[2]\n"
      "sdot z28.s, z6.b, z5.b[2]\n"
      "ld1b { z6.b }, p5/Z, [x14, #-6, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[2]\n"
      "sdot z13.s, z7.b, z1.b[2]\n"
      "sdot z17.s, z7.b, z2.b[2]\n"
      "sdot z21.s, z7.b, z3.b[2]\n"
      "sdot z25.s, z7.b, z4.b[2]\n"
      "sdot z29.s, z7.b, z5.b[2]\n"
      "ld1b { z7.b }, p5/Z, [x14, #-5, MUL VL]\n"
      "sdot z10.s, z6.b, z0.b[2]\n"
      "sdot z14.s, z6.b, z1.b[2]\n"
      "sdot z18.s, z6.b, z2.b[2]\n"
      "sdot z22.s, z6.b, z3.b[2]\n"
      "sdot z26.s, z6.b, z4.b[2]\n"
      "sdot z30.s, z6.b, z5.b[2]\n"
      "ld1b { z6.b }, p5/Z, [x14, #-4, MUL VL]\n"
      "sdot z11.s, z7.b, z0.b[2]\n"
      "sdot z15.s, z7.b, z1.b[2]\n"
      "sdot z19.s, z7.b, z2.b[2]\n"
      "sdot z23.s, z7.b, z3.b[2]\n"
      "sdot z27.s, z7.b, z4.b[2]\n"
      "sdot z31.s, z7.b, z5.b[2]\n"
      "ld1b { z7.b }, p5/Z, [x14, #-3, MUL VL]\n"
      "sdot z8.s, z6.b, z0.b[3]\n"
      "sdot z12.s, z6.b, z1.b[3]\n"
      "sdot z16.s, z6.b, z2.b[3]\n"
      "sdot z20.s, z6.b, z3.b[3]\n"
      "sdot z24.s, z6.b, z4.b[3]\n"
      "sdot z28.s, z6.b, z5.b[3]\n"
      "ld1b { z6.b }, p5/Z, [x14, #-2, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[3]\n"
      "sdot z13.s, z7.b, z1.b[3]\n"
      "sdot z17.s, z7.b, z2.b[3]\n"
      "sdot z21.s, z7.b, z3.b[3]\n"
      "sdot z25.s, z7.b, z4.b[3]\n"
      "sdot z29.s, z7.b, z5.b[3]\n"
      "ld1b { z7.b }, p5/Z, [x14, #-1, MUL VL]\n"
      "sdot z10.s, z6.b, z0.b[3]\n"
      "sdot z14.s, z6.b, z1.b[3]\n"
      "sdot z18.s, z6.b, z2.b[3]\n"
      "sdot z22.s, z6.b, z3.b[3]\n"
      "sdot z26.s, z6.b, z4.b[3]\n"
      "sdot z30.s, z6.b, z5.b[3]\n"
      "sdot z11.s, z7.b, z0.b[3]\n"
      "sdot z15.s, z7.b, z1.b[3]\n"
      "sdot z19.s, z7.b, z2.b[3]\n"
      "sdot z23.s, z7.b, z3.b[3]\n"
      "sdot z27.s, z7.b, z4.b[3]\n"
      "sdot z31.s, z7.b, z5.b[3]\n"
      "bgt 69b\n"
      "70:"  // Height 6: Multiply loop: Single iteration only
      "ld1b { z6.b }, p5/Z, [x14]\n"
      "whilelt p0.b, XZR, x11\n"
      "ld1b { z7.b }, p5/Z, [x14, #1, MUL VL]\n"
      "subs x11, x11, #0x4\n"
      "ld1rqb { z0.b }, p0/Z, [x10]\n"
      "sdot z8.s, z6.b, z0.b[0]\n"
      "ld1rqb { z1.b }, p0/Z, [x28]\n"
      "add x10, x10, #0x10\n"
      "sdot z9.s, z7.b, z0.b[0]\n"
      "ld1rqb { z2.b }, p0/Z, [x26]\n"
      "add x28, x28, #0x10\n"
      "sdot z12.s, z6.b, z1.b[0]\n"
      "ld1rqb { z3.b }, p0/Z, [x24]\n"
      "add x26, x26, #0x10\n"
      "sdot z16.s, z6.b, z2.b[0]\n"
      "ld1rqb { z4.b }, p0/Z, [x22]\n"
      "add x24, x24, #0x10\n"
      "sdot z13.s, z7.b, z1.b[0]\n"
      "ld1rqb { z5.b }, p0/Z, [x20]\n"
      "add x22, x22, #0x10\n"
      "sdot z20.s, z6.b, z3.b[0]\n"
      "add x20, x20, #0x10\n"
      "sdot z17.s, z7.b, z2.b[0]\n"
      "sdot z24.s, z6.b, z4.b[0]\n"
      "sdot z28.s, z6.b, z5.b[0]\n"
      "ld1b { z6.b }, p5/Z, [x14, #2, MUL VL]\n"
      "sdot z21.s, z7.b, z3.b[0]\n"
      "sdot z25.s, z7.b, z4.b[0]\n"
      "sdot z29.s, z7.b, z5.b[0]\n"
      "ld1b { z7.b }, p5/Z, [x14, #3, MUL VL]\n"
      "addvl x14, x14, #4\n"
      "sdot z10.s, z6.b, z0.b[0]\n"
      "sdot z14.s, z6.b, z1.b[0]\n"
      "sdot z18.s, z6.b, z2.b[0]\n"
      "sdot z22.s, z6.b, z3.b[0]\n"
      "sdot z26.s, z6.b, z4.b[0]\n"
      "sdot z30.s, z6.b, z5.b[0]\n"
      "sdot z11.s, z7.b, z0.b[0]\n"
      "sdot z15.s, z7.b, z1.b[0]\n"
      "sdot z19.s, z7.b, z2.b[0]\n"
      "sdot z23.s, z7.b, z3.b[0]\n"
      "sdot z27.s, z7.b, z4.b[0]\n"
      "sdot z31.s, z7.b, z5.b[0]\n"
      "ble 71f\n"
      "ld1b { z6.b }, p5/Z, [x14]\n"
      "sdot z8.s, z6.b, z0.b[1]\n"
      "ld1b { z7.b }, p5/Z, [x14, #1, MUL VL]\n"
      "subs x11, x11, #0x4\n"
      "sdot z12.s, z6.b, z1.b[1]\n"
      "sdot z16.s, z6.b, z2.b[1]\n"
      "sdot z20.s, z6.b, z3.b[1]\n"
      "sdot z24.s, z6.b, z4.b[1]\n"
      "sdot z28.s, z6.b, z5.b[1]\n"
      "ld1b { z6.b }, p5/Z, [x14, #2, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[1]\n"
      "sdot z13.s, z7.b, z1.b[1]\n"
      "sdot z17.s, z7.b, z2.b[1]\n"
      "sdot z21.s, z7.b, z3.b[1]\n"
      "sdot z25.s, z7.b, z4.b[1]\n"
      "sdot z29.s, z7.b, z5.b[1]\n"
      "ld1b { z7.b }, p5/Z, [x14, #3, MUL VL]\n"
      "addvl x14, x14, #4\n"
      "sdot z10.s, z6.b, z0.b[1]\n"
      "sdot z14.s, z6.b, z1.b[1]\n"
      "sdot z18.s, z6.b, z2.b[1]\n"
      "sdot z22.s, z6.b, z3.b[1]\n"
      "sdot z26.s, z6.b, z4.b[1]\n"
      "sdot z30.s, z6.b, z5.b[1]\n"
      "sdot z11.s, z7.b, z0.b[1]\n"
      "sdot z15.s, z7.b, z1.b[1]\n"
      "sdot z19.s, z7.b, z2.b[1]\n"
      "sdot z23.s, z7.b, z3.b[1]\n"
      "sdot z27.s, z7.b, z4.b[1]\n"
      "sdot z31.s, z7.b, z5.b[1]\n"
      "ble 71f\n"
      "ld1b { z6.b }, p5/Z, [x14]\n"
      "sdot z8.s, z6.b, z0.b[2]\n"
      "ld1b { z7.b }, p5/Z, [x14, #1, MUL VL]\n"
      "subs x11, x11, #0x4\n"
      "sdot z12.s, z6.b, z1.b[2]\n"
      "sdot z16.s, z6.b, z2.b[2]\n"
      "sdot z20.s, z6.b, z3.b[2]\n"
      "sdot z24.s, z6.b, z4.b[2]\n"
      "sdot z28.s, z6.b, z5.b[2]\n"
      "ld1b { z6.b }, p5/Z, [x14, #2, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[2]\n"
      "sdot z13.s, z7.b, z1.b[2]\n"
      "sdot z17.s, z7.b, z2.b[2]\n"
      "sdot z21.s, z7.b, z3.b[2]\n"
      "sdot z25.s, z7.b, z4.b[2]\n"
      "sdot z29.s, z7.b, z5.b[2]\n"
      "ld1b { z7.b }, p5/Z, [x14, #3, MUL VL]\n"
      "addvl x14, x14, #4\n"
      "sdot z10.s, z6.b, z0.b[2]\n"
      "sdot z14.s, z6.b, z1.b[2]\n"
      "sdot z18.s, z6.b, z2.b[2]\n"
      "sdot z22.s, z6.b, z3.b[2]\n"
      "sdot z26.s, z6.b, z4.b[2]\n"
      "sdot z30.s, z6.b, z5.b[2]\n"
      "sdot z11.s, z7.b, z0.b[2]\n"
      "sdot z15.s, z7.b, z1.b[2]\n"
      "sdot z19.s, z7.b, z2.b[2]\n"
      "sdot z23.s, z7.b, z3.b[2]\n"
      "sdot z27.s, z7.b, z4.b[2]\n"
      "sdot z31.s, z7.b, z5.b[2]\n"
      "ble 71f\n"
      "ld1b { z6.b }, p5/Z, [x14]\n"
      "sdot z8.s, z6.b, z0.b[3]\n"
      "ld1b { z7.b }, p5/Z, [x14, #1, MUL VL]\n"
      "sdot z12.s, z6.b, z1.b[3]\n"
      "sdot z16.s, z6.b, z2.b[3]\n"
      "sdot z20.s, z6.b, z3.b[3]\n"
      "sdot z24.s, z6.b, z4.b[3]\n"
      "sdot z28.s, z6.b, z5.b[3]\n"
      "ld1b { z6.b }, p5/Z, [x14, #2, MUL VL]\n"
      "sdot z9.s, z7.b, z0.b[3]\n"
      "sdot z13.s, z7.b, z1.b[3]\n"
      "sdot z17.s, z7.b, z2.b[3]\n"
      "sdot z21.s, z7.b, z3.b[3]\n"
      "sdot z25.s, z7.b, z4.b[3]\n"
      "sdot z29.s, z7.b, z5.b[3]\n"
      "ld1b { z7.b }, p5/Z, [x14, #3, MUL VL]\n"
      "addvl x14, x14, #4\n"
      "sdot z10.s, z6.b, z0.b[3]\n"
      "sdot z14.s, z6.b, z1.b[3]\n"
      "sdot z18.s, z6.b, z2.b[3]\n"
      "sdot z22.s, z6.b, z3.b[3]\n"
      "sdot z26.s, z6.b, z4.b[3]\n"
      "sdot z30.s, z6.b, z5.b[3]\n"
      "sdot z11.s, z7.b, z0.b[3]\n"
      "sdot z15.s, z7.b, z1.b[3]\n"
      "sdot z19.s, z7.b, z2.b[3]\n"
      "sdot z23.s, z7.b, z3.b[3]\n"
      "sdot z27.s, z7.b, z4.b[3]\n"
      "sdot z31.s, z7.b, z5.b[3]\n"
      "71:"  // Height 6: Multiply loop: multiply skip
      "prfm pldl1keep, [x10, #0x80]\n"
      "add x12, x12, #0x1\n"
      "prfm pldl1keep, [x28, #0x80]\n"
      "prfm pldl1keep, [x26, #0x80]\n"
      "prfm pldl1keep, [x24, #0x80]\n"
      "prfm pldl1keep, [x22, #0x80]\n"
      "prfm pldl1keep, [x20, #0x80]\n"
      "ldr w19, [%x[args_ptr], %[offsetof_num_strings]]\n"
      "cmp x12, x19\n"
      "bne 66b\n"
      "prfm pstl1keep, [x13, #0x0]\n"
      "prfm pstl1keep, [x9, #0x0]\n"
      "prfm pstl1keep, [x27, #0x0]\n"
      "prfm pstl1keep, [x25, #0x0]\n"
      "prfm pstl1keep, [x23, #0x0]\n"
      "prfm pstl1keep, [x21, #0x0]\n"
      "st1w { z8.s }, p4, [x13]\n"
      "st1w { z9.s }, p3, [x13, #1, MUL VL]\n"
      "st1w { z10.s }, p2, [x13, #2, MUL VL]\n"
      "st1w { z11.s }, p1, [x13, #3, MUL VL]\n"
      "addvl x13, x13, #4\n"
      "st1w { z12.s }, p4, [x9]\n"
      "st1w { z13.s }, p3, [x9, #1, MUL VL]\n"
      "st1w { z14.s }, p2, [x9, #2, MUL VL]\n"
      "st1w { z15.s }, p1, [x9, #3, MUL VL]\n"
      "addvl x9, x9, #4\n"
      "st1w { z16.s }, p4, [x27]\n"
      "st1w { z17.s }, p3, [x27, #1, MUL VL]\n"
      "st1w { z18.s }, p2, [x27, #2, MUL VL]\n"
      "st1w { z19.s }, p1, [x27, #3, MUL VL]\n"
      "addvl x27, x27, #4\n"
      "st1w { z20.s }, p4, [x25]\n"
      "st1w { z21.s }, p3, [x25, #1, MUL VL]\n"
      "st1w { z22.s }, p2, [x25, #2, MUL VL]\n"
      "st1w { z23.s }, p1, [x25, #3, MUL VL]\n"
      "addvl x25, x25, #4\n"
      "st1w { z24.s }, p4, [x23]\n"
      "st1w { z25.s }, p3, [x23, #1, MUL VL]\n"
      "st1w { z26.s }, p2, [x23, #2, MUL VL]\n"
      "st1w { z27.s }, p1, [x23, #3, MUL VL]\n"
      "addvl x23, x23, #4\n"
      "st1w { z28.s }, p4, [x21]\n"
      "st1w { z29.s }, p3, [x21, #1, MUL VL]\n"
      "st1w { z30.s }, p2, [x21, #2, MUL VL]\n"
      "st1w { z31.s }, p1, [x21, #3, MUL VL]\n"
      "addvl x21, x21, #4\n"
      "72:"  // Height 6: Writeback done
      "mov x19, #0x0\n"
      "incw x19, ALL, MUL #4\n"
      "subs x15, x15, x19\n"
      "bgt 63b\n"
      "subs %x[M], %x[M], #0x6\n"
      "beq 74f\n"
      "ldr x20, [%x[args_ptr], %[offsetof_input_offset]]\n"
      "tbz %x[flags], #3, 73f\n"
      "add x20, x20, #0x6\n"
      "str x20, [%x[args_ptr], %[offsetof_input_offset]]\n"
      "b 1b\n"
      "73:"  // Update direct input
      "mov x19, #0x6\n"
      "madd %x[input_ptr], x19, x20, %x[input_ptr]\n"
      "b 1b\n"
      "74:"  // Exit

      : [M] "+r" (M), [input_ptr] "+r" (input_ptr), [output_ptr] "+r" (output_ptr)
      : [args_ptr] "r" (&ka), [flags] "r" (flags), [offsetof_B_ptr] "I" (offsetof(KernelArgs, B_ptr)), [offsetof_N] "I" (offsetof(KernelArgs, N)), [offsetof_input_initial_col] "I" (offsetof(KernelArgs, input_initial_col)), [offsetof_input_offset] "I" (offsetof(KernelArgs, input_offset)), [offsetof_num_strings] "I" (offsetof(KernelArgs, num_strings)), [offsetof_output_offset] "I" (offsetof(KernelArgs, output_offset)), [offsetof_string_lengths] "I" (offsetof(KernelArgs, string_lengths))
      : "cc", "memory", "p0", "p1", "p2", "p3", "p4", "p5", "x9", "x10", "x11", "x12", "x13", "x14", "x15", "x19", "x20", "x21", "x22", "x23", "x24", "x25", "x26", "x27", "x28", "z0", "z1", "z2", "z3", "z4", "z5", "z6", "z7", "z8", "z9", "z10", "z11", "z12", "z13", "z14", "z15", "z16", "z17", "z18", "z19", "z20", "z21", "z22", "z23", "z24", "z25", "z26", "z27", "z28", "z29", "z30", "z31"
    );
}

} // namespace arm_gemm
#endif // __ARM_FEATURE_SVE
